@@ -25,15 +25,10 @@
 #include <grub/fs.h>
 #include <grub/file.h>
 #include <grub/procfs.h>
+#include <grub/partition.h>
 
 #ifdef GRUB_UTIL
-#include <errno.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
 #include <grub/emu/hostdisk.h>
-#include <unistd.h>
-#include <string.h>
 #endif
 
 GRUB_MOD_LICENSE ("GPLv3+");
@@ -57,7 +52,7 @@ static inline int GF_PER_SECTOR (const struct grub_cryptodisk *dev)
 }
 
 static grub_cryptodisk_t cryptodisk_list = NULL;
-static grub_uint8_t n = 0;
+static grub_uint8_t last_cryptodisk_id = 0;
 
 static void
 gf_mul_x (grub_uint8_t *g)
@@ -115,19 +110,24 @@ grub_crypto_pcbc_decrypt (grub_crypto_cipher_handle_t cipher,
 			 void *iv)
 {
   grub_uint8_t *inptr, *outptr, *end;
-  grub_uint8_t ivt[cipher->cipher->blocksize];
+  grub_uint8_t ivt[GRUB_CRYPTO_MAX_CIPHER_BLOCKSIZE];
+  grub_size_t blocksize;
   if (!cipher->cipher->decrypt)
     return GPG_ERR_NOT_SUPPORTED;
-  if (size % cipher->cipher->blocksize != 0)
+  blocksize = cipher->cipher->blocksize;
+  if (blocksize == 0 || (((blocksize - 1) & blocksize) != 0)
+      || ((size & (blocksize - 1)) != 0))
+    return GPG_ERR_INV_ARG;
+  if (blocksize > GRUB_CRYPTO_MAX_CIPHER_BLOCKSIZE)
     return GPG_ERR_INV_ARG;
   end = (grub_uint8_t *) in + size;
   for (inptr = in, outptr = out; inptr < end;
-       inptr += cipher->cipher->blocksize, outptr += cipher->cipher->blocksize)
+       inptr += blocksize, outptr += blocksize)
     {
-      grub_memcpy (ivt, inptr, cipher->cipher->blocksize);
+      grub_memcpy (ivt, inptr, blocksize);
       cipher->cipher->decrypt (cipher->ctx, outptr, inptr);
-      grub_crypto_xor (outptr, outptr, iv, cipher->cipher->blocksize);
-      grub_crypto_xor (iv, ivt, outptr, cipher->cipher->blocksize);
+      grub_crypto_xor (outptr, outptr, iv, blocksize);
+      grub_crypto_xor (iv, ivt, outptr, blocksize);
     }
   return GPG_ERR_NO_ERROR;
 }
@@ -138,19 +138,24 @@ grub_crypto_pcbc_encrypt (grub_crypto_cipher_handle_t cipher,
 			  void *iv)
 {
   grub_uint8_t *inptr, *outptr, *end;
-  grub_uint8_t ivt[cipher->cipher->blocksize];
-  if (!cipher->cipher->decrypt)
+  grub_uint8_t ivt[GRUB_CRYPTO_MAX_CIPHER_BLOCKSIZE];
+  grub_size_t blocksize;
+  if (!cipher->cipher->encrypt)
     return GPG_ERR_NOT_SUPPORTED;
-  if (size % cipher->cipher->blocksize != 0)
+  blocksize = cipher->cipher->blocksize;
+  if (blocksize > GRUB_CRYPTO_MAX_CIPHER_BLOCKSIZE)
+    return GPG_ERR_INV_ARG;
+  if (blocksize == 0 || (((blocksize - 1) & blocksize) != 0)
+      || ((size & (blocksize - 1)) != 0))
     return GPG_ERR_INV_ARG;
   end = (grub_uint8_t *) in + size;
   for (inptr = in, outptr = out; inptr < end;
-       inptr += cipher->cipher->blocksize, outptr += cipher->cipher->blocksize)
+       inptr += blocksize, outptr += blocksize)
     {
-      grub_memcpy (ivt, inptr, cipher->cipher->blocksize);
-      grub_crypto_xor (outptr, outptr, iv, cipher->cipher->blocksize);
+      grub_memcpy (ivt, inptr, blocksize);
+      grub_crypto_xor (outptr, outptr, iv, blocksize);
       cipher->cipher->encrypt (cipher->ctx, outptr, inptr);
-      grub_crypto_xor (iv, ivt, outptr, cipher->cipher->blocksize);
+      grub_crypto_xor (iv, ivt, outptr, blocksize);
     }
   return GPG_ERR_NO_ERROR;
 }
@@ -224,6 +229,9 @@ grub_cryptodisk_endecrypt (struct grub_cryptodisk *dev,
   grub_size_t i;
   gcry_err_code_t err;
 
+  if (dev->cipher->cipher->blocksize > GRUB_CRYPTO_MAX_CIPHER_BLOCKSIZE)
+    return GPG_ERR_INV_ARG;
+
   /* The only mode without IV.  */
   if (dev->mode == GRUB_CRYPTODISK_MODE_ECB && !dev->rekey)
     return (do_encrypt ? grub_crypto_ecb_encrypt (dev->cipher, data, data, len)
@@ -234,7 +242,7 @@ grub_cryptodisk_endecrypt (struct grub_cryptodisk *dev,
       grub_size_t sz = ((dev->cipher->cipher->blocksize
 			 + sizeof (grub_uint32_t) - 1)
 			/ sizeof (grub_uint32_t));
-      grub_uint32_t iv[sz];
+      grub_uint32_t iv[(GRUB_CRYPTO_MAX_CIPHER_BLOCKSIZE + 3) / 4];
 
       if (dev->rekey)
 	{
@@ -248,7 +256,7 @@ grub_cryptodisk_endecrypt (struct grub_cryptodisk *dev,
 	    }
 	}
 
-      grub_memset (iv, 0, sz * sizeof (iv[0]));
+      grub_memset (iv, 0, sizeof (iv));
       switch (dev->mode_iv)
 	{
 	case GRUB_CRYPTODISK_MODE_IV_NULL:
@@ -256,9 +264,11 @@ grub_cryptodisk_endecrypt (struct grub_cryptodisk *dev,
 	case GRUB_CRYPTODISK_MODE_IV_BYTECOUNT64_HASH:
 	  {
 	    grub_uint64_t tmp;
-	    GRUB_PROPERLY_ALIGNED_ARRAY (ctx, dev->iv_hash->contextsize);
+	    void *ctx;
 
-	    grub_memset (ctx, 0, sizeof (ctx));
+	    ctx = grub_zalloc (dev->iv_hash->contextsize);
+	    if (!ctx)
+	      return GPG_ERR_OUT_OF_MEMORY;
 
 	    tmp = grub_cpu_to_le64 (sector << dev->log_sector_size);
 	    dev->iv_hash->init (ctx);
@@ -267,10 +277,12 @@ grub_cryptodisk_endecrypt (struct grub_cryptodisk *dev,
 	    dev->iv_hash->final (ctx);
 
 	    grub_memcpy (iv, dev->iv_hash->read (ctx), sizeof (iv));
+	    grub_free (ctx);
 	  }
 	  break;
 	case GRUB_CRYPTODISK_MODE_IV_PLAIN64:
 	  iv[1] = grub_cpu_to_le32 (sector >> 32);
+	  /* FALLTHROUGH */
 	case GRUB_CRYPTODISK_MODE_IV_PLAIN:
 	  iv[0] = grub_cpu_to_le32 (sector & 0xFFFFFFFF);
 	  break;
@@ -368,11 +380,13 @@ grub_cryptodisk_endecrypt (struct grub_cryptodisk *dev,
 	  break;
 	case GRUB_CRYPTODISK_MODE_ECB:
 	  if (do_encrypt)
-	    grub_crypto_ecb_encrypt (dev->cipher, data + i, data + i,
-				     (1U << dev->log_sector_size));
+	    err = grub_crypto_ecb_encrypt (dev->cipher, data + i, data + i,
+					   (1U << dev->log_sector_size));
 	  else
-	    grub_crypto_ecb_decrypt (dev->cipher, data + i, data + i,
-				     (1U << dev->log_sector_size));
+	    err = grub_crypto_ecb_decrypt (dev->cipher, data + i, data + i,
+					   (1U << dev->log_sector_size));
+	  if (err)
+	    return err;
 	  break;
 	default:
 	  return GPG_ERR_NOT_IMPLEMENTED;
@@ -413,7 +427,9 @@ grub_cryptodisk_setkey (grub_cryptodisk_t dev, grub_uint8_t *key, grub_size_t ke
   if (dev->mode_iv == GRUB_CRYPTODISK_MODE_IV_ESSIV)
     {
       grub_size_t essiv_keysize = dev->essiv_hash->mdlen;
-      grub_uint8_t hashed_key[essiv_keysize];
+      grub_uint8_t hashed_key[GRUB_CRYPTO_MAX_MDLEN];
+      if (essiv_keysize > GRUB_CRYPTO_MAX_MDLEN)
+	return GPG_ERR_INV_ARG;
 
       grub_crypto_hash (dev->essiv_hash, hashed_key, key, keysize);
       err = grub_crypto_cipher_set_key (dev->essiv_cipher,
@@ -504,11 +520,11 @@ grub_cryptodisk_open (const char *name, grub_disk_t disk)
 #ifdef GRUB_UTIL
   if (dev->cheat)
     {
-      if (dev->cheat_fd == -1)
-	dev->cheat_fd = open (dev->cheat, O_RDONLY);
-      if (dev->cheat_fd == -1)
+      if (!GRUB_UTIL_FD_IS_VALID (dev->cheat_fd))
+	dev->cheat_fd = grub_util_fd_open (dev->cheat, GRUB_UTIL_FD_O_RDONLY);
+      if (!GRUB_UTIL_FD_IS_VALID (dev->cheat_fd))
 	return grub_error (GRUB_ERR_IO, N_("cannot open `%s': %s"),
-			   dev->cheat, strerror (errno));
+			   dev->cheat, grub_util_fd_strerror ());
     }
 #endif
 
@@ -523,6 +539,7 @@ grub_cryptodisk_open (const char *name, grub_disk_t disk)
 
   disk->data = dev;
   disk->total_sectors = dev->total_length;
+  disk->max_agglomerate = GRUB_DISK_MAX_MAX_AGGLOMERATE;
   disk->id = dev->id;
   dev->ref++;
   return GRUB_ERR_NONE;
@@ -541,8 +558,8 @@ grub_cryptodisk_close (grub_disk_t disk)
 #ifdef GRUB_UTIL
   if (dev->cheat)
     {
-      close (dev->cheat_fd);
-      dev->cheat_fd = -1;
+      grub_util_fd_close (dev->cheat_fd);
+      dev->cheat_fd = GRUB_UTIL_FD_INVALID;
     }
 #endif
   grub_disk_close (dev->source_disk);
@@ -560,14 +577,15 @@ grub_cryptodisk_read (grub_disk_t disk, grub_disk_addr_t sector,
 #ifdef GRUB_UTIL
   if (dev->cheat)
     {
-      err = grub_util_fd_seek (dev->cheat_fd, dev->cheat,
-			       sector << disk->log_sector_size);
-      if (err)
-	return err;
+      int r;
+      r = grub_util_fd_seek (dev->cheat_fd, sector << disk->log_sector_size);
+      if (r)
+	return grub_error (GRUB_ERR_BAD_DEVICE, N_("cannot seek `%s': %s"),
+			   dev->cheat, grub_util_fd_strerror ());
       if (grub_util_fd_read (dev->cheat_fd, buf, size << disk->log_sector_size)
 	  != (ssize_t) (size << disk->log_sector_size))
 	return grub_error (GRUB_ERR_READ_ERROR, N_("cannot read `%s': %s"),
-			   dev->cheat, strerror (errno));
+			   dev->cheat, grub_util_fd_strerror ());
       return GRUB_ERR_NONE;
     }
 #endif
@@ -604,14 +622,15 @@ grub_cryptodisk_write (grub_disk_t disk, grub_disk_addr_t sector,
 #ifdef GRUB_UTIL
   if (dev->cheat)
     {
-      err = grub_util_fd_seek (dev->cheat_fd, dev->cheat,
-			       sector << disk->log_sector_size);
-      if (err)
-	return err;
+      int r;
+      r = grub_util_fd_seek (dev->cheat_fd, sector << disk->log_sector_size);
+      if (r)
+	return grub_error (GRUB_ERR_BAD_DEVICE, N_("cannot seek `%s': %s"),
+			   dev->cheat, grub_util_fd_strerror ());
       if (grub_util_fd_write (dev->cheat_fd, buf, size << disk->log_sector_size)
 	  != (ssize_t) (size << disk->log_sector_size))
 	return grub_error (GRUB_ERR_READ_ERROR, N_("cannot read `%s': %s"),
-			   dev->cheat, strerror (errno));
+			   dev->cheat, grub_util_fd_strerror ());
       return GRUB_ERR_NONE;
     }
 #endif
@@ -635,10 +654,16 @@ grub_cryptodisk_write (grub_disk_t disk, grub_disk_addr_t sector,
       return grub_crypto_gcry_error (gcry_err);
     }
 
-  err = grub_disk_write (dev->source_disk,
-			  (sector << (disk->log_sector_size
-				      - GRUB_DISK_SECTOR_BITS)) + dev->offset,
-			  0, size << disk->log_sector_size, tmp);
+  /* Since ->write was called so disk.mod is loaded but be paranoid  */
+  
+  if (grub_disk_write_weak)
+    err = grub_disk_write_weak (dev->source_disk,
+				(sector << (disk->log_sector_size
+					    - GRUB_DISK_SECTOR_BITS))
+				+ dev->offset,
+				0, size << disk->log_sector_size, tmp);
+  else
+    err = grub_error (GRUB_ERR_BUG, "disk.mod not loaded");
   grub_free (tmp);
   return err;
 }
@@ -692,9 +717,10 @@ grub_cryptodisk_insert (grub_cryptodisk_t newdev, const char *name,
       return grub_errno;
     }
 
-  newdev->id = n++;
+  newdev->id = last_cryptodisk_id++;
   newdev->source_id = source->id;
   newdev->source_dev_id = source->dev->id;
+  newdev->partition_start = grub_partition_get_start (source->partition);
   newdev->next = cryptodisk_list;
   cryptodisk_list = newdev;
 
@@ -717,7 +743,9 @@ grub_cryptodisk_get_by_source_disk (grub_disk_t disk)
   grub_cryptodisk_t dev;
   for (dev = cryptodisk_list; dev != NULL; dev = dev->next)
     if (dev->source_id == disk->id && dev->source_dev_id == disk->dev->id)
-      return dev;
+      if ((disk->partition && grub_partition_get_start (disk->partition) == dev->partition_start) ||
+          (!disk->partition && dev->partition_start == 0))
+        return dev;
   return NULL;
 }
 
@@ -735,10 +763,11 @@ grub_cryptodisk_cheat_insert (grub_cryptodisk_t newdev, const char *name,
       return grub_errno;
     }
 
-  newdev->cheat_fd = -1;
+  newdev->cheat_fd = GRUB_UTIL_FD_INVALID;
   newdev->source_id = source->id;
   newdev->source_dev_id = source->dev->id;
-  newdev->id = n++;
+  newdev->partition_start = grub_partition_get_start (source->partition);
+  newdev->id = last_cryptodisk_id++;
   newdev->next = cryptodisk_list;
   cryptodisk_list = newdev;
 
@@ -746,31 +775,34 @@ grub_cryptodisk_cheat_insert (grub_cryptodisk_t newdev, const char *name,
 }
 
 void
-grub_util_cryptodisk_print_abstraction (grub_disk_t disk)
+grub_util_cryptodisk_get_abstraction (grub_disk_t disk,
+				      void (*cb) (const char *val, void *data),
+				      void *data)
 {
   grub_cryptodisk_t dev = (grub_cryptodisk_t) disk->data;
 
-  grub_printf ("cryptodisk %s ", dev->modname);
+  cb ("cryptodisk", data);
+  cb (dev->modname, data);
 
   if (dev->cipher)
-    grub_printf ("%s ", dev->cipher->cipher->modname);
+    cb (dev->cipher->cipher->modname, data);
   if (dev->secondary_cipher)
-    grub_printf ("%s ", dev->secondary_cipher->cipher->modname);
+    cb (dev->secondary_cipher->cipher->modname, data);
   if (dev->essiv_cipher)
-    grub_printf ("%s ", dev->essiv_cipher->cipher->modname);
+    cb (dev->essiv_cipher->cipher->modname, data);
   if (dev->hash)
-    grub_printf ("%s ", dev->hash->modname);
+    cb (dev->hash->modname, data);
   if (dev->essiv_hash)
-    grub_printf ("%s ", dev->essiv_hash->modname);
+    cb (dev->essiv_hash->modname, data);
   if (dev->iv_hash)
-    grub_printf ("%s ", dev->iv_hash->modname);
+    cb (dev->iv_hash->modname, data);
 }
 
-void
-grub_util_cryptodisk_print_uuid (grub_disk_t disk)
+const char *
+grub_util_cryptodisk_get_uuid (grub_disk_t disk)
 {
   grub_cryptodisk_t dev = (grub_cryptodisk_t) disk->data;
-  grub_printf ("%s ", dev->uuid);
+  return dev->uuid;
 }
 
 #endif
@@ -938,33 +970,43 @@ grub_cmd_cryptomount (grub_extcmd_context_t ctxt, int argc, char **args)
       grub_disk_t disk;
       grub_cryptodisk_t dev;
       char *diskname;
-      char *disklast;
+      char *disklast = NULL;
+      grub_size_t len;
 
       search_uuid = NULL;
       check_boot = state[2].set;
       diskname = args[0];
-      if (diskname[0] == '(' && *(disklast = &diskname[grub_strlen (diskname) - 1]) == ')')
+      len = grub_strlen (diskname);
+      if (len && diskname[0] == '(' && diskname[len - 1] == ')')
 	{
+	  disklast = &diskname[len - 1];
 	  *disklast = '\0';
-	  disk = grub_disk_open (diskname + 1);
-	  *disklast = ')';
+	  diskname++;
 	}
-      else
-	disk = grub_disk_open (diskname);
+
+      disk = grub_disk_open (diskname);
       if (!disk)
-	return grub_errno;
+	{
+	  if (disklast)
+	    *disklast = ')';
+	  return grub_errno;
+	}
 
       dev = grub_cryptodisk_get_by_source_disk (disk);
       if (dev)
 	{
 	  grub_dprintf ("cryptodisk", "already mounted as crypto%lu\n", dev->id);
 	  grub_disk_close (disk);
+	  if (disklast)
+	    *disklast = ')';
 	  return GRUB_ERR_NONE;
 	}
 
-      err = grub_cryptodisk_scan_device_real (args[0], disk);
+      err = grub_cryptodisk_scan_device_real (diskname, disk);
 
       grub_disk_close (disk);
+      if (disklast)
+	*disklast = ')';
 
       return err;
     }
@@ -973,13 +1015,13 @@ grub_cmd_cryptomount (grub_extcmd_context_t ctxt, int argc, char **args)
 static struct grub_disk_dev grub_cryptodisk_dev = {
   .name = "cryptodisk",
   .id = GRUB_DISK_DEVICE_CRYPTODISK_ID,
-  .iterate = grub_cryptodisk_iterate,
-  .open = grub_cryptodisk_open,
-  .close = grub_cryptodisk_close,
-  .read = grub_cryptodisk_read,
-  .write = grub_cryptodisk_write,
+  .disk_iterate = grub_cryptodisk_iterate,
+  .disk_open = grub_cryptodisk_open,
+  .disk_close = grub_cryptodisk_close,
+  .disk_read = grub_cryptodisk_read,
+  .disk_write = grub_cryptodisk_write,
 #ifdef GRUB_UTIL
-  .memberlist = grub_cryptodisk_memberlist,
+  .disk_memberlist = grub_cryptodisk_memberlist,
 #endif
   .next = 0
 };
@@ -994,12 +1036,14 @@ hex (grub_uint8_t val)
 
 /* Open a file named NAME and initialize FILE.  */
 static char *
-luks_script_get (void)
+luks_script_get (grub_size_t *sz)
 {
   grub_cryptodisk_t i;
   grub_size_t size = 0;
   char *ptr, *ret;
 
+  *sz = 0;
+
   for (i = cryptodisk_list; i != NULL; i = i->next)
     if (grub_strcmp (i->modname, "luks") == 0)
       {
@@ -1081,6 +1125,7 @@ luks_script_get (void)
 	*ptr++ = '\n';
       }
   *ptr = '\0';
+  *sz = ptr - ret;
   return ret;
 }
 
