@@ -328,11 +328,11 @@ static struct grub_disk_dev grub_cbfsdisk_dev =
   {
     .name = "cbfsdisk",
     .id = GRUB_DISK_DEVICE_CBFSDISK_ID,
-    .iterate = grub_cbfsdisk_iterate,
-    .open = grub_cbfsdisk_open,
-    .close = grub_cbfsdisk_close,
-    .read = grub_cbfsdisk_read,
-    .write = grub_cbfsdisk_write,
+    .disk_iterate = grub_cbfsdisk_iterate,
+    .disk_open = grub_cbfsdisk_open,
+    .disk_close = grub_cbfsdisk_close,
+    .disk_read = grub_cbfsdisk_read,
+    .disk_write = grub_cbfsdisk_write,
     .next = 0
   };
 
@@ -375,10 +375,10 @@ fini_cbfsdisk (void)
 
 static struct grub_fs grub_cbfs_fs = {
   .name = "cbfs",
-  .dir = grub_cbfs_dir,
-  .open = grub_cbfs_open,
-  .read = grub_cbfs_read,
-  .close = grub_cbfs_close,
+  .fs_dir = grub_cbfs_dir,
+  .fs_open = grub_cbfs_open,
+  .fs_read = grub_cbfs_read,
+  .fs_close = grub_cbfs_close,
 #ifdef GRUB_UTIL
   .reserved_first_sector = 0,
   .blocklist_install = 0,
