@@ -1417,13 +1417,13 @@ grub_hfs_uuid (grub_device_t device, char **uuid)
 static struct grub_fs grub_hfs_fs =
   {
     .name = "hfs",
-    .dir = grub_hfs_dir,
-    .open = grub_hfs_open,
-    .read = grub_hfs_read,
-    .close = grub_hfs_close,
-    .label = grub_hfs_label,
-    .uuid = grub_hfs_uuid,
-    .mtime = grub_hfs_mtime,
+    .fs_dir = grub_hfs_dir,
+    .fs_open = grub_hfs_open,
+    .fs_read = grub_hfs_read,
+    .fs_close = grub_hfs_close,
+    .fs_label = grub_hfs_label,
+    .fs_uuid = grub_hfs_uuid,
+    .fs_mtime = grub_hfs_mtime,
 #ifdef GRUB_UTIL
     .reserved_first_sector = 1,
     .blocklist_install = 1,
