@@ -87,7 +87,7 @@ fixup (grub_uint8_t *buf, grub_size_t len, const grub_uint8_t *magic)
 }
 
 static grub_err_t read_mft (struct grub_ntfs_data *data, grub_uint8_t *buf,
-			    grub_uint32_t mftno);
+			    grub_uint64_t mftno);
 static grub_err_t read_attr (struct grub_ntfs_attr *at, grub_uint8_t *dest,
 			     grub_disk_addr_t ofs, grub_size_t len,
 			     int cached,
@@ -283,25 +283,17 @@ locate_attr (struct grub_ntfs_attr *at, struct grub_ntfs_file *mft,
   return pa;
 }
 
-static grub_uint8_t *
-read_run_data (grub_uint8_t *run, int nn, grub_disk_addr_t * val, int sig)
+static grub_disk_addr_t
+read_run_data (const grub_uint8_t *run, int nn, int sig)
 {
-  grub_disk_addr_t r, v;
+  grub_uint64_t r = 0;
 
-  r = 0;
-  v = 1;
+  if (sig && nn && (run[nn - 1] & 0x80))
+    r = -1;
 
-  while (nn--)
-    {
-      r += v * (*(run++));
-      v <<= 8;
-    }
+  grub_memcpy (&r, run, nn);
 
-  if ((sig) && (r & (v >> 1)))
-    r -= v;
-
-  *val = r;
-  return run;
+  return grub_le_to_cpu64 (r);
 }
 
 grub_err_t
@@ -313,8 +305,9 @@ grub_ntfs_read_run_list (struct grub_ntfs_rlst * ctx)
 
   run = ctx->cur_run;
 retry:
-  c1 = ((*run) & 0xF);
-  c2 = ((*run) >> 4);
+  c1 = ((*run) & 0x7);
+  c2 = ((*run) >> 4) & 0x7;
+  run++;
   if (!c1)
     {
       if ((ctx->attr) && (ctx->attr->flags & GRUB_NTFS_AF_ALST))
@@ -338,10 +331,11 @@ retry:
 	}
       return grub_error (GRUB_ERR_BAD_FS, "run list overflown");
     }
-  run = read_run_data (run + 1, c1, &val, 0);	/* length of current VCN */
   ctx->curr_vcn = ctx->next_vcn;
-  ctx->next_vcn += val;
-  run = read_run_data (run, c2, &val, 1);	/* offset to previous LCN */
+  ctx->next_vcn += read_run_data (run, c1, 0);	/* length of current VCN */
+  run += c1;
+  val = read_run_data (run, c2, 1);	/* offset to previous LCN */
+  run += c2;
   ctx->curr_lcn += val;
   if (val == 0)
     ctx->flags |= GRUB_NTFS_RF_BLNK;
@@ -373,7 +367,6 @@ read_data (struct grub_ntfs_attr *at, grub_uint8_t *pa, grub_uint8_t *dest,
 	   grub_disk_addr_t ofs, grub_size_t len, int cached,
 	   grub_disk_read_hook_t read_hook, void *read_hook_data)
 {
-  grub_disk_addr_t vcn;
   struct grub_ntfs_rlst cc, *ctx;
 
   if (len == 0)
@@ -385,6 +378,9 @@ read_data (struct grub_ntfs_attr *at, grub_uint8_t *pa, grub_uint8_t *dest,
   ctx->comp.log_spc = at->mft->data->log_spc;
   ctx->comp.disk = at->mft->data->disk;
 
+  if (read_hook == grub_file_progress_hook)
+    ctx->file = read_hook_data;
+
   if (pa[8] == 0)
     {
       if (ofs + len > u32at (pa, 0x10))
@@ -393,52 +389,23 @@ read_data (struct grub_ntfs_attr *at, grub_uint8_t *pa, grub_uint8_t *dest,
       return 0;
     }
 
-  if (u16at (pa, 0xC) & GRUB_NTFS_FLAG_COMPRESSED)
-    ctx->flags |= GRUB_NTFS_RF_COMP;
-  else
-    ctx->flags &= ~GRUB_NTFS_RF_COMP;
   ctx->cur_run = pa + u16at (pa, 0x20);
 
-  if (ctx->flags & GRUB_NTFS_RF_COMP)
+  ctx->next_vcn = u32at (pa, 0x10);
+  ctx->curr_lcn = 0;
+
+  if ((pa[0xC] & GRUB_NTFS_FLAG_COMPRESSED)
+      && !(at->flags & GRUB_NTFS_AF_GPOS))
     {
       if (!cached)
 	return grub_error (GRUB_ERR_BAD_FS, "attribute can\'t be compressed");
 
-      if (at->sbuf)
-	{
-	  if ((ofs & (~(GRUB_NTFS_COM_LEN - 1))) == at->save_pos)
-	    {
-	      grub_disk_addr_t n;
-
-	      n = GRUB_NTFS_COM_LEN - (ofs - at->save_pos);
-	      if (n > len)
-		n = len;
-
-	      grub_memcpy (dest, at->sbuf + ofs - at->save_pos, n);
-	      if (n == len)
-		return 0;
-
-	      dest += n;
-	      len -= n;
-	      ofs += n;
-	    }
-	}
-      else
-	{
-	  at->sbuf = grub_malloc (GRUB_NTFS_COM_LEN);
-	  if (at->sbuf == NULL)
-	    return grub_errno;
-	  at->save_pos = 1;
-	}
-
-      vcn = ctx->target_vcn = (ofs >> GRUB_NTFS_COM_LOG_LEN) * (GRUB_NTFS_COM_SEC >> ctx->comp.log_spc);
-      ctx->target_vcn &= ~0xFULL;
+      return (grub_ntfscomp_func) ? grub_ntfscomp_func (dest, ofs, len, ctx)
+	: grub_error (GRUB_ERR_BAD_FS, N_("module `%s' isn't loaded"),
+		      "ntfscomp");
     }
-  else
-    vcn = ctx->target_vcn = ofs >> (GRUB_NTFS_BLK_SHR + ctx->comp.log_spc);
 
-  ctx->next_vcn = u32at (pa, 0x10);
-  ctx->curr_lcn = 0;
+  ctx->target_vcn = ofs >> (GRUB_NTFS_BLK_SHR + ctx->comp.log_spc);
   while (ctx->next_vcn <= ctx->target_vcn)
     {
       if (grub_ntfs_read_run_list (ctx))
@@ -467,20 +434,12 @@ read_data (struct grub_ntfs_attr *at, grub_uint8_t *pa, grub_uint8_t *dest,
       return 0;
     }
 
-  if (!(ctx->flags & GRUB_NTFS_RF_COMP))
-    {
-      grub_fshelp_read_file (ctx->comp.disk, (grub_fshelp_node_t) ctx,
-			     read_hook, read_hook_data, ofs, len,
-			     (char *) dest,
-			     grub_ntfs_read_block, ofs + len,
-			     ctx->comp.log_spc, 0);
-      return grub_errno;
-    }
-
-  return (grub_ntfscomp_func) ? grub_ntfscomp_func (at, dest, ofs, len, ctx,
-						    vcn) :
-    grub_error (GRUB_ERR_BAD_FS, N_("module `%s' isn't loaded"),
-		"ntfscomp");
+  grub_fshelp_read_file (ctx->comp.disk, (grub_fshelp_node_t) ctx,
+			 read_hook, read_hook_data, ofs, len,
+			 (char *) dest,
+			 grub_ntfs_read_block, ofs + len,
+			 ctx->comp.log_spc, 0);
+  return grub_errno;
 }
 
 static grub_err_t
@@ -532,17 +491,17 @@ read_attr (struct grub_ntfs_attr *at, grub_uint8_t *dest, grub_disk_addr_t ofs,
 }
 
 static grub_err_t
-read_mft (struct grub_ntfs_data *data, grub_uint8_t *buf, grub_uint32_t mftno)
+read_mft (struct grub_ntfs_data *data, grub_uint8_t *buf, grub_uint64_t mftno)
 {
   if (read_attr
       (&data->mmft.attr, buf, mftno * ((grub_disk_addr_t) data->mft_size << GRUB_NTFS_BLK_SHR),
        data->mft_size << GRUB_NTFS_BLK_SHR, 0, 0, 0))
-    return grub_error (GRUB_ERR_BAD_FS, "read MFT 0x%X fails", mftno);
+    return grub_error (GRUB_ERR_BAD_FS, "read MFT 0x%llx fails", (unsigned long long) mftno);
   return fixup (buf, data->mft_size, (const grub_uint8_t *) "FILE");
 }
 
 static grub_err_t
-init_file (struct grub_ntfs_file *mft, grub_uint32_t mftno)
+init_file (struct grub_ntfs_file *mft, grub_uint64_t mftno)
 {
   unsigned short flag;
 
@@ -557,7 +516,8 @@ init_file (struct grub_ntfs_file *mft, grub_uint32_t mftno)
 
   flag = u16at (mft->buf, 0x16);
   if ((flag & 1) == 0)
-    return grub_error (GRUB_ERR_BAD_FS, "MFT 0x%X is not in use", mftno);
+    return grub_error (GRUB_ERR_BAD_FS, "MFT 0x%llx is not in use",
+		       (unsigned long long) mftno);
 
   if ((flag & 2) == 0)
     {
@@ -565,7 +525,8 @@ init_file (struct grub_ntfs_file *mft, grub_uint32_t mftno)
 
       pa = locate_attr (&mft->attr, mft, GRUB_NTFS_AT_DATA);
       if (pa == NULL)
-	return grub_error (GRUB_ERR_BAD_FS, "no $DATA in MFT 0x%X", mftno);
+	return grub_error (GRUB_ERR_BAD_FS, "no $DATA in MFT 0x%llx",
+			   (unsigned long long) mftno);
 
       if (!pa[8])
 	mft->size = u32at (pa, 0x10);
@@ -588,6 +549,28 @@ free_file (struct grub_ntfs_file *mft)
   grub_free (mft->buf);
 }
 
+static char *
+get_utf8 (grub_uint8_t *in, grub_size_t len)
+{
+  grub_uint8_t *buf;
+  grub_uint16_t *tmp;
+  grub_size_t i;
+
+  buf = grub_malloc (len * GRUB_MAX_UTF8_PER_UTF16 + 1);
+  tmp = grub_malloc (len * sizeof (tmp[0]));
+  if (!buf || !tmp)
+    {
+      grub_free (buf);
+      grub_free (tmp);
+      return NULL;
+    }
+  for (i = 0; i < len; i++)
+    tmp[i] = grub_le_to_cpu16 (grub_get_unaligned16 (in + 2 * i));
+  *grub_utf16_to_utf8 (buf, tmp, len) = '\0';
+  grub_free (tmp);
+  return (char *) buf;
+}
+
 static int
 list_file (struct grub_ntfs_file *diro, grub_uint8_t *pos,
 	   grub_fshelp_iterate_dir_hook_t hook, void *hook_data)
@@ -617,12 +600,6 @@ list_file (struct grub_ntfs_file *diro, grub_uint8_t *pos,
 	  struct grub_ntfs_file *fdiro;
 	  grub_uint32_t attr;
 
-	  if (u16at (pos, 4))
-	    {
-	      grub_error (GRUB_ERR_BAD_FS, "64-bit MFT number");
-	      return 0;
-	    }
-
 	  attr = u32at (pos, 0x48);
 	  if (attr & GRUB_NTFS_ATTR_REPARSE)
 	    type = GRUB_FSHELP_SYMLINK;
@@ -636,21 +613,15 @@ list_file (struct grub_ntfs_file *diro, grub_uint8_t *pos,
 	    return 0;
 
 	  fdiro->data = diro->data;
-	  fdiro->ino = u32at (pos, 0);
+	  fdiro->ino = u64at (pos, 0) & 0xffffffffffffULL;
 	  fdiro->mtime = u64at (pos, 0x20);
 
-	  ustr = grub_malloc (ns * GRUB_MAX_UTF8_PER_UTF16 + 1);
+	  ustr = get_utf8 (np, ns);
 	  if (ustr == NULL)
-	    return 0;
-	  {
-	    grub_uint16_t tmp[ns];
-	    int i;
-	    for (i = 0; i < ns; i++)
-	      tmp[i] = grub_le_to_cpu16 (grub_get_unaligned16 ((char *) np
-							       + 2 * i));
-
-	    *grub_utf16_to_utf8 ((grub_uint8_t *) ustr, tmp, ns) = '\0';
-	  }
+	    {
+	      grub_free (fdiro);
+	      return 0;
+	    }
           if (namespace)
             type |= GRUB_FSHELP_CASE_INSENSITIVE;
 
@@ -675,7 +646,7 @@ struct symlink_descriptor
   grub_uint16_t len1;
   grub_uint16_t off2;
   grub_uint16_t len2;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 static char *
 grub_ntfs_read_symlink (grub_fshelp_node_t node)
@@ -683,10 +654,9 @@ grub_ntfs_read_symlink (grub_fshelp_node_t node)
   struct grub_ntfs_file *mft;
   struct symlink_descriptor symdesc;
   grub_err_t err;
-  grub_uint16_t *buf16;
+  grub_uint8_t *buf16;
   char *buf, *end;
   grub_size_t len;
-  grub_size_t i;
   grub_uint8_t *pa;
   grub_size_t off;
 
@@ -702,7 +672,8 @@ grub_ntfs_read_symlink (grub_fshelp_node_t node)
   pa = locate_attr (&mft->attr, mft, GRUB_NTFS_AT_SYMLINK);
   if (pa == NULL)
     {
-      grub_error (GRUB_ERR_BAD_FS, "no $SYMLINK in MFT 0x%X", mft->ino);
+      grub_error (GRUB_ERR_BAD_FS, "no $SYMLINK in MFT 0x%llx",
+		  (unsigned long long) mft->ino);
       return NULL;
     }
 
@@ -733,26 +704,22 @@ grub_ntfs_read_symlink (grub_fshelp_node_t node)
   if (!buf16)
     return NULL;
 
-  err = read_attr (&mft->attr, (grub_uint8_t *) buf16, off, len, 1, 0, 0);
+  err = read_attr (&mft->attr, buf16, off, len, 1, 0, 0);
   if (err)
     return NULL;
 
-  buf = grub_malloc (len * 2 + 1);
+  buf = get_utf8 (buf16, len / 2);
   if (!buf)
     {
       grub_free (buf16);
       return NULL;
     }
+  grub_free (buf16);
 
-  for (i = 0; i < len / 2; i++)
-    {
-      buf16[i] = grub_le_to_cpu16 (buf16[i]);
-      if (buf16[i] == '\\')
-	buf16[i] = '/';
-    }
+  for (end = buf; *end; end++)
+    if (*end == '\\')
+      *end = '/';
 
-  end = (char *) grub_utf16_to_utf8 ((grub_uint8_t *) buf, buf16, len / 2);
-  *end = '\0';
   /* Split the sequence to avoid GCC thinking that this is a trigraph.  */
   if (grub_memcmp (buf, "/?" "?/", 4) == 0 && buf[5] == ':' && buf[6] == '/'
       && grub_isalpha (buf[4]))
@@ -876,7 +843,8 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
 
   if (bitmap)
     {
-      grub_disk_addr_t v, i;
+      grub_disk_addr_t i;
+      grub_uint8_t v;
 
       indx = grub_malloc (mft->data->idx_size << GRUB_NTFS_BLK_SHR);
       if (indx == NULL)
@@ -899,7 +867,7 @@ grub_ntfs_iterate_dir (grub_fshelp_node_t dir,
 		goto done;
 	    }
 	  v <<= 1;
-	  if (v >= 0x100)
+	  if (!v)
 	    {
 	      v = 1;
 	      bitmap++;
@@ -945,19 +913,23 @@ grub_ntfs_mount (grub_disk_t disk)
   spc = (((grub_uint32_t) bpb.sectors_per_cluster
 	  * (grub_uint32_t) grub_le_to_cpu16 (bpb.bytes_per_sector))
 	 >> GRUB_NTFS_BLK_SHR);
-  if (spc == 0 || (spc & (spc - 1)))
+  if (spc == 0)
     goto fail;
 
   for (data->log_spc = 0; (1U << data->log_spc) < spc; data->log_spc++);
 
   if (bpb.clusters_per_mft > 0)
     data->mft_size = ((grub_disk_addr_t) bpb.clusters_per_mft) << data->log_spc;
+  else if (-bpb.clusters_per_mft < GRUB_NTFS_BLK_SHR || -bpb.clusters_per_mft >= 31)
+    goto fail;
   else
     data->mft_size = 1ULL << (-bpb.clusters_per_mft - GRUB_NTFS_BLK_SHR);
 
   if (bpb.clusters_per_index > 0)
     data->idx_size = (((grub_disk_addr_t) bpb.clusters_per_index)
 		      << data->log_spc);
+  else if (-bpb.clusters_per_index < GRUB_NTFS_BLK_SHR || -bpb.clusters_per_index >= 31)
+    goto fail;
   else
     data->idx_size = 1ULL << (-bpb.clusters_per_index - GRUB_NTFS_BLK_SHR);
 
@@ -1184,20 +1156,11 @@ grub_ntfs_label (grub_device_t device, char **label)
   pa = find_attr (&mft->attr, GRUB_NTFS_AT_VOLUME_NAME);
   if ((pa) && (pa[8] == 0) && (u32at (pa, 0x10)))
     {
-      grub_uint8_t *buf;
       int len;
 
       len = u32at (pa, 0x10) / 2;
-      buf = grub_malloc (len * 4 + 1);
       pa += u16at (pa, 0x14);
-      {
-	grub_uint16_t tmp[len];
-	int i;
-	for (i = 0; i < len; i++)
-	  tmp[i] = grub_le_to_cpu16 (grub_get_unaligned16 (pa + 2 * i));
-	*grub_utf16_to_utf8 (buf, tmp, len) = '\0';
-      }
-      *label = (char *) buf;
+      *label = get_utf8 (pa, len);
     }
 
 fail:
@@ -1249,12 +1212,12 @@ grub_ntfs_uuid (grub_device_t device, char **uuid)
 static struct grub_fs grub_ntfs_fs =
   {
     .name = "ntfs",
-    .dir = grub_ntfs_dir,
-    .open = grub_ntfs_open,
-    .read = grub_ntfs_read,
-    .close = grub_ntfs_close,
-    .label = grub_ntfs_label,
-    .uuid = grub_ntfs_uuid,
+    .fs_dir = grub_ntfs_dir,
+    .fs_open = grub_ntfs_open,
+    .fs_read = grub_ntfs_read,
+    .fs_close = grub_ntfs_close,
+    .fs_label = grub_ntfs_label,
+    .fs_uuid = grub_ntfs_uuid,
 #ifdef GRUB_UTIL
     .reserved_first_sector = 1,
     .blocklist_install = 1,
