@@ -104,7 +104,7 @@ decomp_block (struct grub_ntfs_comp *cc, grub_uint8_t *dest)
 	      if (tag & 1)
 		{
 		  grub_uint32_t i, len, delta, code, lmask, dshift;
-		  grub_uint16_t word;
+		  grub_uint16_t word = 0;
 
 		  if (decomp_get16 (cc, &word))
 		    return grub_errno;
@@ -226,6 +226,9 @@ read_block (struct grub_ntfs_rlst *ctx, grub_uint8_t *buf, grub_size_t num)
 		{
 		  grub_memset (buf, 0, nn * GRUB_NTFS_COM_LEN);
 		  buf += nn * GRUB_NTFS_COM_LEN;
+		  if (grub_file_progress_hook && ctx->file)
+		    grub_file_progress_hook (0, 0, nn * GRUB_NTFS_COM_LEN,
+					     ctx->file);
 		}
 	    }
 	  else
@@ -236,6 +239,9 @@ read_block (struct grub_ntfs_rlst *ctx, grub_uint8_t *buf, grub_size_t num)
 		    return grub_errno;
 		  if (buf)
 		    buf += GRUB_NTFS_COM_LEN;
+		  if (grub_file_progress_hook && ctx->file)
+		    grub_file_progress_hook (0, 0, GRUB_NTFS_COM_LEN,
+					     ctx->file);
 		  nn--;
 		}
 	    }
@@ -262,6 +268,11 @@ read_block (struct grub_ntfs_rlst *ctx, grub_uint8_t *buf, grub_size_t num)
 			 ctx->target_vcn)) << ctx->comp.log_spc, 0,
 		       tt << (ctx->comp.log_spc + GRUB_NTFS_BLK_SHR), buf))
 		    return grub_errno;
+		  if (grub_file_progress_hook && ctx->file)
+		    grub_file_progress_hook (0, 0,
+					     tt << (ctx->comp.log_spc
+						    + GRUB_NTFS_BLK_SHR),
+					     ctx->file);
 		  buf += tt << (ctx->comp.log_spc + GRUB_NTFS_BLK_SHR);
 		}
 	      nn -= tt;
@@ -280,6 +291,11 @@ read_block (struct grub_ntfs_rlst *ctx, grub_uint8_t *buf, grub_size_t num)
 		       nn << (ctx->comp.log_spc + GRUB_NTFS_BLK_SHR), buf))
 		    return grub_errno;
 		  buf += nn << (ctx->comp.log_spc + GRUB_NTFS_BLK_SHR);
+		  if (grub_file_progress_hook && ctx->file)
+		    grub_file_progress_hook (0, 0,
+					     nn << (ctx->comp.log_spc
+						    + GRUB_NTFS_BLK_SHR),
+					     ctx->file);
 		}
 	      ctx->target_vcn += nn;
 	    }
@@ -289,10 +305,48 @@ read_block (struct grub_ntfs_rlst *ctx, grub_uint8_t *buf, grub_size_t num)
 }
 
 static grub_err_t
-ntfscomp (struct grub_ntfs_attr *at, grub_uint8_t *dest, grub_disk_addr_t ofs,
-	  grub_size_t len, struct grub_ntfs_rlst *ctx, grub_disk_addr_t vcn)
+ntfscomp (grub_uint8_t *dest, grub_disk_addr_t ofs,
+	  grub_size_t len, struct grub_ntfs_rlst *ctx)
 {
   grub_err_t ret;
+  grub_disk_addr_t vcn;
+
+  if (ctx->attr->sbuf)
+    {
+      if ((ofs & (~(GRUB_NTFS_COM_LEN - 1))) == ctx->attr->save_pos)
+	{
+	  grub_disk_addr_t n;
+
+	  n = GRUB_NTFS_COM_LEN - (ofs - ctx->attr->save_pos);
+	  if (n > len)
+	    n = len;
+
+	  grub_memcpy (dest, ctx->attr->sbuf + ofs - ctx->attr->save_pos, n);
+	  if (grub_file_progress_hook && ctx->file)
+	    grub_file_progress_hook (0, 0, n, ctx->file);
+	  if (n == len)
+	    return 0;
+
+	  dest += n;
+	  len -= n;
+	  ofs += n;
+	}
+    }
+  else
+    {
+      ctx->attr->sbuf = grub_malloc (GRUB_NTFS_COM_LEN);
+      if (ctx->attr->sbuf == NULL)
+	return grub_errno;
+      ctx->attr->save_pos = 1;
+    }
+
+  vcn = ctx->target_vcn = (ofs >> GRUB_NTFS_COM_LOG_LEN) * (GRUB_NTFS_COM_SEC >> ctx->comp.log_spc);
+  ctx->target_vcn &= ~0xFULL;
+  while (ctx->next_vcn <= ctx->target_vcn)
+    {
+      if (grub_ntfs_read_run_list (ctx))
+	return grub_errno;
+    }
 
   ctx->comp.comp_head = ctx->comp.comp_tail = 0;
   ctx->comp.cbuf = grub_malloc (1 << (ctx->comp.log_spc + GRUB_NTFS_BLK_SHR));
@@ -315,21 +369,28 @@ ntfscomp (struct grub_ntfs_attr *at, grub_uint8_t *dest, grub_disk_addr_t ofs,
   if (ofs % GRUB_NTFS_COM_LEN)
     {
       grub_uint32_t t, n, o;
+      void *file = ctx->file;
+
+      ctx->file = 0;
 
       t = ctx->target_vcn << (ctx->comp.log_spc + GRUB_NTFS_BLK_SHR);
-      if (read_block (ctx, at->sbuf, 1))
+      if (read_block (ctx, ctx->attr->sbuf, 1))
 	{
 	  ret = grub_errno;
 	  goto quit;
 	}
 
-      at->save_pos = t;
+      ctx->file = file;
+
+      ctx->attr->save_pos = t;
 
       o = ofs % GRUB_NTFS_COM_LEN;
       n = GRUB_NTFS_COM_LEN - o;
       if (n > len)
 	n = len;
-      grub_memcpy (dest, &at->sbuf[o], n);
+      grub_memcpy (dest, &ctx->attr->sbuf[o], n);
+      if (grub_file_progress_hook && ctx->file)
+	grub_file_progress_hook (0, 0, n, ctx->file);
       if (n == len)
 	goto quit;
       dest += n;
@@ -347,17 +408,21 @@ ntfscomp (struct grub_ntfs_attr *at, grub_uint8_t *dest, grub_disk_addr_t ofs,
   if (len)
     {
       grub_uint32_t t;
+      void *file = ctx->file;
 
+      ctx->file = 0;
       t = ctx->target_vcn << (ctx->comp.log_spc + GRUB_NTFS_BLK_SHR);
-      if (read_block (ctx, at->sbuf, 1))
+      if (read_block (ctx, ctx->attr->sbuf, 1))
 	{
 	  ret = grub_errno;
 	  goto quit;
 	}
 
-      at->save_pos = t;
+      ctx->attr->save_pos = t;
 
-      grub_memcpy (dest, at->sbuf, len);
+      grub_memcpy (dest, ctx->attr->sbuf, len);
+      if (grub_file_progress_hook && file)
+	grub_file_progress_hook (0, 0, len, file);
     }
 
 quit:
