@@ -35,7 +35,7 @@ struct grub_sfs_bheader
   grub_uint8_t magic[4];
   grub_uint32_t chksum;
   grub_uint32_t ipointtomyself;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* The sfs rootblock.  */
 struct grub_sfs_rblock
@@ -50,7 +50,7 @@ struct grub_sfs_rblock
   grub_uint8_t unused3[8];
   grub_uint32_t rootobject;
   grub_uint32_t btree;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 enum
   {
@@ -69,18 +69,18 @@ struct grub_sfs_obj
     {
       grub_uint32_t first_block;
       grub_uint32_t size;
-    } file __attribute__ ((packed));
+    } GRUB_PACKED file;
     struct
     {
       grub_uint32_t hashtable;
       grub_uint32_t dir_objc;
-    } dir __attribute__ ((packed));
+    } GRUB_PACKED dir;
   } file_dir;
   grub_uint32_t mtime;
   grub_uint8_t type;
   grub_uint8_t filename[1];
   grub_uint8_t comment[1];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 #define	GRUB_SFS_TYPE_DELETED	32
 #define	GRUB_SFS_TYPE_SYMLINK	64
@@ -95,13 +95,13 @@ struct grub_sfs_objc
   grub_uint32_t prev;
   /* The amount of objects depends on the blocksize.  */
   struct grub_sfs_obj objects[1];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_sfs_btree_node
 {
   grub_uint32_t key;
   grub_uint32_t data;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_sfs_btree_extent
 {
@@ -109,7 +109,7 @@ struct grub_sfs_btree_extent
   grub_uint32_t next;
   grub_uint32_t prev;
   grub_uint16_t size;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_sfs_btree
 {
@@ -120,7 +120,7 @@ struct grub_sfs_btree
   /* Normally this can be kind of node, but just extents are
      supported.  */
   struct grub_sfs_btree_node node[1];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 
 
@@ -173,10 +173,11 @@ grub_sfs_read_extent (struct grub_sfs_data *data, unsigned int block,
   struct grub_sfs_btree *tree;
   int i;
   grub_uint32_t next;
+  grub_size_t blocksize = GRUB_DISK_SECTOR_SIZE << data->log_blocksize;
 
-  treeblock = grub_malloc (GRUB_DISK_SECTOR_SIZE << data->log_blocksize);
-  if (!block)
-    return 0;
+  treeblock = grub_malloc (blocksize);
+  if (!treeblock)
+    return grub_errno;
 
   next = grub_be_to_cpu32 (data->rblock.btree);
   tree = (struct grub_sfs_btree *) treeblock;
@@ -184,17 +185,21 @@ grub_sfs_read_extent (struct grub_sfs_data *data, unsigned int block,
   /* Handle this level in the btree.  */
   do
     {
+      grub_uint16_t nnodes;
       grub_disk_read (data->disk,
 		      ((grub_disk_addr_t) next) << data->log_blocksize,
-		      0, GRUB_DISK_SECTOR_SIZE << data->log_blocksize,
-		      treeblock);
+		      0, blocksize, treeblock);
       if (grub_errno)
 	{
 	  grub_free (treeblock);
 	  return grub_errno;
 	}
 
-      for (i = grub_be_to_cpu16 (tree->nodes) - 1; i >= 0; i--)
+      nnodes = grub_be_to_cpu16 (tree->nodes);
+      if (nnodes * (grub_uint32_t) (tree)->nodesize > blocksize)
+	break;
+
+      for (i = (int) nnodes - 1; i >= 0; i--)
 	{
 
 #define EXTNODE(tree, index)						\
@@ -736,11 +741,11 @@ grub_sfs_label (grub_device_t device, char **label)
 static struct grub_fs grub_sfs_fs =
   {
     .name = "sfs",
-    .dir = grub_sfs_dir,
-    .open = grub_sfs_open,
-    .read = grub_sfs_read,
-    .close = grub_sfs_close,
-    .label = grub_sfs_label,
+    .fs_dir = grub_sfs_dir,
+    .fs_open = grub_sfs_open,
+    .fs_read = grub_sfs_read,
+    .fs_close = grub_sfs_close,
+    .fs_label = grub_sfs_label,
 #ifdef GRUB_UTIL
     .reserved_first_sector = 0,
     .blocklist_install = 1,
