@@ -32,6 +32,7 @@
 #include <grub/env.h>
 #include <grub/cache.h>
 #include <grub/time.h>
+#include <grub/cpu/cpuid.h>
 #include <grub/cpu/tsc.h>
 #include <grub/machine/time.h>
 
@@ -161,13 +162,13 @@ mmap_iterate_hook (grub_uint64_t addr, grub_uint64_t size,
 		   void *data __attribute__ ((unused)))
 {
   /* Avoid the lower memory.  */
-  if (addr < 0x100000)
+  if (addr < GRUB_MEMORY_MACHINE_UPPER_START)
     {
-      if (size <= 0x100000 - addr)
+      if (size <= GRUB_MEMORY_MACHINE_UPPER_START - addr)
 	return 0;
 
-      size -= 0x100000 - addr;
-      addr = 0x100000;
+      size -= GRUB_MEMORY_MACHINE_UPPER_START - addr;
+      addr = GRUB_MEMORY_MACHINE_UPPER_START;
     }
 
   /* Ignore >4GB.  */
@@ -184,6 +185,26 @@ mmap_iterate_hook (grub_uint64_t addr, grub_uint64_t size,
   return 0;
 }
 
+extern grub_uint16_t grub_bios_via_workaround1, grub_bios_via_workaround2;
+
+/* Via needs additional wbinvd.  */
+static void
+grub_via_workaround_init (void)
+{
+  grub_uint32_t manufacturer[3], max_cpuid;
+  if (! grub_cpu_is_cpuid_supported ())
+    return;
+
+  grub_cpuid (0, max_cpuid, manufacturer[0], manufacturer[2], manufacturer[1]);
+
+  if (grub_memcmp (manufacturer, "CentaurHauls", 12) != 0)
+    return;
+
+  grub_bios_via_workaround1 = 0x090f;
+  grub_bios_via_workaround2 = 0x090f;
+  asm volatile ("wbinvd");
+}
+
 void
 grub_machine_init (void)
 {
@@ -193,6 +214,9 @@ grub_machine_init (void)
 #endif
   grub_addr_t modend;
 
+  /* This has to happen before any BIOS calls. */
+  grub_via_workaround_init ();
+
   grub_modbase = GRUB_MEMORY_MACHINE_DECOMPRESSION_ADDR + (_edata - _start);
 
   /* Initialize the console as early as possible.  */
@@ -239,8 +263,9 @@ grub_machine_init (void)
 }
 
 void
-grub_machine_fini (void)
+grub_machine_fini (int flags)
 {
-  grub_console_fini ();
+  if (flags & GRUB_LOADER_FLAG_NORETURN)
+    grub_console_fini ();
   grub_stop_floppy ();
 }
