@@ -25,8 +25,39 @@
 #include <grub/env.h>
 #include <grub/i18n.h>
 
+union printf_arg
+{
+  /* Yes, type is also part of union as the moment we fill the value
+     we don't need to store its type anymore (when we'll need it, we'll
+     have format spec again. So save some space.  */
+  enum
+    {
+      INT, LONG, LONGLONG,
+      UNSIGNED_INT = 3, UNSIGNED_LONG, UNSIGNED_LONGLONG
+    } type;
+  long long ll;
+};
+
+struct printf_args
+{
+  union printf_arg prealloc[32];
+  union printf_arg *ptr;
+  grub_size_t count;
+};
+
+static void
+parse_printf_args (const char *fmt0, struct printf_args *args,
+		   va_list args_in);
 static int
-grub_vsnprintf_real (char *str, grub_size_t n, const char *fmt, va_list args);
+grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0,
+		     struct printf_args *args);
+
+static void
+free_printf_args (struct printf_args *args)
+{
+  if (args->ptr != args->prealloc)
+    grub_free (args->ptr);
+}
 
 static int
 grub_iswordseparator (int c)
@@ -64,25 +95,6 @@ grub_memmove (void *dest, const void *src, grub_size_t n)
   return dest;
 }
 
-#ifndef __APPLE__
-void *memmove (void *dest, const void *src, grub_size_t n)
-  __attribute__ ((alias ("grub_memmove")));
-/* GCC emits references to memcpy() for struct copies etc.  */
-void *memcpy (void *dest, const void *src, grub_size_t n)
-  __attribute__ ((alias ("grub_memmove")));
-#else
-void * __attribute__ ((regparm(0)))
-memcpy (void *dest, const void *src, grub_size_t n)
-{
-	return grub_memmove (dest, src, n);
-}
-void * __attribute__ ((regparm(0)))
-memmove (void *dest, const void *src, grub_size_t n)
-{
-	return grub_memmove (dest, src, n);
-}
-#endif
-
 char *
 grub_strcpy (char *dest, const char *src)
 {
@@ -94,17 +106,6 @@ grub_strcpy (char *dest, const char *src)
   return dest;
 }
 
-char *
-grub_strncpy (char *dest, const char *src, int c)
-{
-  char *p = dest;
-
-  while ((*p++ = *src++) != '\0' && --c)
-    ;
-
-  return dest;
-}
-
 int
 grub_printf (const char *fmt, ...)
 {
@@ -180,15 +181,16 @@ grub_real_dprintf (const char *file, const int line, const char *condition,
 #define PREALLOC_SIZE 255
 
 int
-grub_vprintf (const char *fmt, va_list args)
+grub_vprintf (const char *fmt, va_list ap)
 {
   grub_size_t s;
   static char buf[PREALLOC_SIZE + 1];
   char *curbuf = buf;
-  va_list ap2;
-  va_copy (ap2, args);
+  struct printf_args args;
 
-  s = grub_vsnprintf_real (buf, PREALLOC_SIZE, fmt, args);
+  parse_printf_args (fmt, &args, ap);
+
+  s = grub_vsnprintf_real (buf, PREALLOC_SIZE, fmt, &args);
   if (s > PREALLOC_SIZE)
     {
       curbuf = grub_malloc (s + 1);
@@ -202,16 +204,16 @@ grub_vprintf (const char *fmt, va_list args)
 	  curbuf = buf;
 	}
       else
-	s = grub_vsnprintf_real (curbuf, s, fmt, ap2);
+	s = grub_vsnprintf_real (curbuf, s, fmt, &args);
     }
 
-  va_end (ap2);
+  free_printf_args (&args);
 
   grub_xputs (curbuf);
 
   if (curbuf != buf)
     grub_free (curbuf);
-  
+
   return s;
 }
 
@@ -232,16 +234,6 @@ grub_memcmp (const void *s1, const void *s2, grub_size_t n)
 
   return 0;
 }
-#ifndef __APPLE__
-int memcmp (const void *s1, const void *s2, grub_size_t n)
-  __attribute__ ((alias ("grub_memcmp")));
-#else
-int __attribute__ ((regparm(0)))
-memcmp (const void *s1, const void *s2, grub_size_t n)
-{
-  return grub_memcmp (s1, s2, n);
-}
-#endif
 
 int
 grub_strcmp (const char *s1, const char *s2)
@@ -347,13 +339,6 @@ grub_isspace (int c)
   return (c == '\n' || c == '\r' || c == ' ' || c == '\t');
 }
 
-int
-grub_isprint (int c)
-{
-  return (c >= ' ' && c <= '~');
-}
-
-
 unsigned long
 grub_strtoul (const char *str, char **end, int base)
 {
@@ -378,7 +363,8 @@ grub_strtoull (const char *str, char **end, int base)
   int found = 0;
 
   /* Skip white spaces.  */
-  while (*str && grub_isspace (*str))
+  /* grub_isspace checks that *str != '\0'.  */
+  while (grub_isspace (*str))
     str++;
 
   /* Guess the base, if not specified. The prefix `0x' means 16, and
@@ -405,12 +391,13 @@ grub_strtoull (const char *str, char **end, int base)
       unsigned long digit;
 
       digit = grub_tolower (*str) - '0';
-      if (digit > 9)
-	{
-	  digit += '0' - 'a' + 10;
-	  if (digit >= (unsigned long) base)
-	    break;
-	}
+      if (digit >= 'a' - '0')
+	digit += '0' - 'a' + 10;
+      else if (digit > 9)
+	break;
+
+      if (digit >= (unsigned long) base)
+	break;
 
       found = 1;
 
@@ -471,6 +458,15 @@ grub_strndup (const char *s, grub_size_t n)
   return p;
 }
 
+/* clang detects that we're implementing here a memset so it decides to
+   optimise and calls memset resulting in infinite recursion. With volatile
+   we make it not optimise in this way.  */
+#ifdef __clang__
+#define VOLATILE_CLANG volatile
+#else
+#define VOLATILE_CLANG
+#endif
+
 void *
 grub_memset (void *s, int c, grub_size_t len)
 {
@@ -487,13 +483,13 @@ grub_memset (void *s, int c, grub_size_t len)
 
       while (len > 0 && (((grub_addr_t) p) & (sizeof (unsigned long) - 1)))
 	{
-	  *(grub_uint8_t *) p = pattern8;
+	  *(VOLATILE_CLANG grub_uint8_t *) p = pattern8;
 	  p = (grub_uint8_t *) p + 1;
 	  len--;
 	}
       while (len >= sizeof (unsigned long))
 	{
-	  *(unsigned long *) p = patternl;
+	  *(VOLATILE_CLANG unsigned long *) p = patternl;
 	  p = (unsigned long *) p + 1;
 	  len -= sizeof (unsigned long);
 	}
@@ -501,23 +497,13 @@ grub_memset (void *s, int c, grub_size_t len)
 
   while (len > 0)
     {
-      *(grub_uint8_t *) p = pattern8;
+      *(VOLATILE_CLANG grub_uint8_t *) p = pattern8;
       p = (grub_uint8_t *) p + 1;
       len--;
     }
 
   return s;
 }
-#ifndef __APPLE__
-void *memset (void *s, int c, grub_size_t n)
-  __attribute__ ((alias ("grub_memset")));
-#else
-void * __attribute__ ((regparm(0)))
-memset (void *s, int c, grub_size_t n)
-{
-  return grub_memset (s, c, n);
-}
-#endif
 
 grub_size_t
 grub_strlen (const char *s)
@@ -559,6 +545,11 @@ grub_divmod64 (grub_uint64_t n, grub_uint64_t d, grub_uint64_t *r)
   grub_uint64_t q = 0;
   grub_uint64_t m = 0;
 
+  /* ARM and IA64 don't have a fast 32-bit division.
+     Using that code would just make us use software division routines, calling
+     ourselves indirectly and hence getting infinite recursion.
+  */
+#if !GRUB_DIVISION_IN_SOFTWARE
   /* Skip the slow computation if 32-bit arithmetic is possible.  */
   if (n < 0xffffffff && d < 0xffffffff)
     {
@@ -567,6 +558,7 @@ grub_divmod64 (grub_uint64_t n, grub_uint64_t d, grub_uint64_t *r)
 
       return ((grub_uint32_t) n) / (grub_uint32_t) d;
     }
+#endif
 
   while (bits--)
     {
@@ -593,7 +585,7 @@ grub_divmod64 (grub_uint64_t n, grub_uint64_t d, grub_uint64_t *r)
 
 /* Convert a long long value to a string. This function avoids 64-bit
    modular arithmetic or divisions.  */
-static char *
+static inline char *
 grub_lltoa (char *str, int c, unsigned long long n)
 {
   unsigned base = (c == 'x') ? 16 : 10;
@@ -631,38 +623,21 @@ grub_lltoa (char *str, int c, unsigned long long n)
   return p;
 }
 
-static int
-grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0, va_list args_in)
+static void
+parse_printf_args (const char *fmt0, struct printf_args *args,
+		   va_list args_in)
 {
+  const char *fmt;
   char c;
   grub_size_t n = 0;
-  grub_size_t count = 0;
-  grub_size_t count_args = 0;
-  const char *fmt;
-  auto void write_char (unsigned char ch);
-  auto void write_str (const char *s);
-  auto void write_fill (const char ch, int n);
 
-  void write_char (unsigned char ch)
-    {
-      if (count < max_len)
-	*str++ = ch;
+  args->count = 0;
 
-      count++;
-    }
-
-  void write_str (const char *s)
-    {
-      while (*s)
-	write_char (*s++);
-    }
-
-  void write_fill (const char ch, int count_fill)
-    {
-      int i;
-      for (i = 0; i < count_fill; i++)
-	write_char (ch);
-    }
+  COMPILE_TIME_ASSERT (sizeof (int) == sizeof (grub_uint32_t));
+  COMPILE_TIME_ASSERT (sizeof (int) <= sizeof (long long));
+  COMPILE_TIME_ASSERT (sizeof (long) <= sizeof (long long));
+  COMPILE_TIME_ASSERT (sizeof (long long) == sizeof (void *)
+		       || sizeof (int) == sizeof (void *));
 
   fmt = fmt0;
   while ((c = *fmt++) != 0)
@@ -670,34 +645,33 @@ grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0, va_list a
       if (c != '%')
 	continue;
 
-      if (*fmt && *fmt =='-')
+      if (*fmt =='-')
 	fmt++;
 
-      while (*fmt && grub_isdigit (*fmt))
+      while (grub_isdigit (*fmt))
 	fmt++;
 
-      if (*fmt && *fmt == '$')
+      if (*fmt == '$')
 	fmt++;
 
-      if (*fmt && *fmt =='-')
+      if (*fmt =='-')
 	fmt++;
 
-      while (*fmt && grub_isdigit (*fmt))
+      while (grub_isdigit (*fmt))
 	fmt++;
 
-      if (*fmt && *fmt =='.')
+      if (*fmt =='.')
 	fmt++;
 
-      while (*fmt && grub_isdigit (*fmt))
+      while (grub_isdigit (*fmt))
 	fmt++;
 
       c = *fmt++;
       if (c == 'l')
-	{
-	  c = *fmt++;
-	  if (c == 'l')
-	    c = *fmt++;
-	}
+	c = *fmt++;
+      if (c == 'l')
+	c = *fmt++;
+
       switch (c)
 	{
 	case 'p':
@@ -707,29 +681,31 @@ grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0, va_list a
 	case 'c':
 	case 'C':
 	case 's':
-	  count_args++;
+	  args->count++;
 	  break;
 	}
     }
 
-  enum { INT, WCHAR, LONG, LONGLONG, POINTER } types[count_args];
-  union 
-  { 
-    int i;
-    grub_uint32_t w;
-    long l;
-    long long ll;
-    void *p;
-  } args[count_args];
+  if (args->count <= ARRAY_SIZE (args->prealloc))
+    args->ptr = args->prealloc;
+  else
+    {
+      args->ptr = grub_malloc (args->count * sizeof (args->ptr[0]));
+      if (!args->ptr)
+	{
+	  grub_errno = GRUB_ERR_NONE;
+	  args->ptr = args->prealloc;
+	  args->count = ARRAY_SIZE (args->prealloc);
+	}
+    }
 
-  grub_memset (types, 0, sizeof (types));
+  grub_memset (args->ptr, 0, args->count * sizeof (args->ptr[0]));
 
   fmt = fmt0;
   n = 0;
   while ((c = *fmt++) != 0)
     {
       int longfmt = 0;
-      int longlongfmt = 0;
       grub_size_t curn;
       const char *p;
 
@@ -738,109 +714,127 @@ grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0, va_list a
 
       curn = n++;
 
-      if (*fmt && *fmt =='-')
+      if (*fmt =='-')
 	fmt++;
 
       p = fmt;
 
-      while (*fmt && grub_isdigit (*fmt))
+      while (grub_isdigit (*fmt))
 	fmt++;
 
-      if (*fmt && *fmt == '$')
+      if (*fmt == '$')
 	{
 	  curn = grub_strtoull (p, 0, 10) - 1;
 	  fmt++;
 	}
 
-      if (*fmt && *fmt =='-')
+      if (*fmt =='-')
 	fmt++;
 
-      while (*fmt && grub_isdigit (*fmt))
+      while (grub_isdigit (*fmt))
 	fmt++;
 
-      if (*fmt && *fmt =='.')
+      if (*fmt =='.')
 	fmt++;
 
-      while (*fmt && grub_isdigit (*fmt))
+      while (grub_isdigit (*fmt))
 	fmt++;
 
       c = *fmt++;
+      if (c == '%')
+	{
+	  n--;
+	  continue;
+	}
+
       if (c == 'l')
 	{
 	  c = *fmt++;
 	  longfmt = 1;
-	  if (c == 'l')
-	    {
-	      c = *fmt++;
-	      longlongfmt = 1;
-	    }
 	}
-      if (curn >= count_args)
+      if (c == 'l')
+	{
+	  c = *fmt++;
+	  longfmt = 2;
+	}
+      if (curn >= args->count)
 	continue;
       switch (c)
 	{
 	case 'x':
 	case 'u':
+	  args->ptr[curn].type = UNSIGNED_INT + longfmt;
+	  break;
 	case 'd':
-	  if (longlongfmt)
-	    types[curn] = LONGLONG;
-	  else if (longfmt)
-	    types[curn] = LONG;
-	  else
-	    types[curn] = INT;
+	  args->ptr[curn].type = INT + longfmt;
 	  break;
 	case 'p':
 	case 's':
-	  types[curn] = POINTER;
-	  break;
-	case 'c':
-	  types[curn] = INT;
+	  if (sizeof (void *) == sizeof (long long))
+	    args->ptr[curn].type = UNSIGNED_LONGLONG;
+	  else
+	    args->ptr[curn].type = UNSIGNED_INT;
 	  break;
 	case 'C':
-	  types[curn] = WCHAR;
+	case 'c':
+	  args->ptr[curn].type = INT;
 	  break;
 	}
     }
 
-  for (n = 0; n < count_args; n++)
-    switch (types[n])
+  for (n = 0; n < args->count; n++)
+    switch (args->ptr[n].type)
       {
-      case WCHAR:
-	args[n].w = va_arg (args_in, grub_uint32_t);
-	break;
-      case POINTER:
-	args[n].p = va_arg (args_in, void *);
-	break;
       case INT:
-	args[n].i = va_arg (args_in, int);
+	args->ptr[n].ll = va_arg (args_in, int);
 	break;
       case LONG:
-	args[n].l = va_arg (args_in, long);
+	args->ptr[n].ll = va_arg (args_in, long);
+	break;
+      case UNSIGNED_INT:
+	args->ptr[n].ll = va_arg (args_in, unsigned int);
+	break;
+      case UNSIGNED_LONG:
+	args->ptr[n].ll = va_arg (args_in, unsigned long);
 	break;
       case LONGLONG:
-	args[n].ll = va_arg (args_in, long long);
+      case UNSIGNED_LONGLONG:
+	args->ptr[n].ll = va_arg (args_in, long long);
 	break;
       }
+}
+
+static inline void __attribute__ ((always_inline))
+write_char (char *str, grub_size_t *count, grub_size_t max_len, unsigned char ch)
+{
+  if (*count < max_len)
+    str[*count] = ch;
+
+  (*count)++;
+}
+
+static int
+grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0,
+		     struct printf_args *args)
+{
+  char c;
+  grub_size_t n = 0;
+  grub_size_t count = 0;
+  const char *fmt;
 
   fmt = fmt0;
 
-  n = 0;
   while ((c = *fmt++) != 0)
     {
-      char tmp[32];
-      char *p;
       unsigned int format1 = 0;
       unsigned int format2 = ~ 0U;
       char zerofill = ' ';
-      int rightfill = 0;
-      int longfmt = 0;
-      int longlongfmt = 0;
-      int unsig = 0;
+      char rightfill = 0;
       grub_size_t curn;
-      
+
       if (c != '%')
 	{
-	  write_char (c);
+	  write_char (str, &count, max_len,c);
 	  continue;
 	}
 
@@ -848,44 +842,26 @@ grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0, va_list a
 
     rescan:;
 
-      if (*fmt && *fmt =='-')
+      if (*fmt =='-')
 	{
 	  rightfill = 1;
 	  fmt++;
 	}
 
-      p = (char *) fmt;
       /* Read formatting parameters.  */
-      while (*p && grub_isdigit (*p))
-	p++;
-
-      if (p > fmt)
+      if (grub_isdigit (*fmt))
 	{
-	  char s[p - fmt + 1];
-	  grub_strncpy (s, fmt, p - fmt);
-	  s[p - fmt] = 0;
-	  if (s[0] == '0')
+	  if (fmt[0] == '0')
 	    zerofill = '0';
-	  format1 = grub_strtoul (s, 0, 10);
-	  fmt = p;
+	  format1 = grub_strtoul (fmt, (char **) &fmt, 10);
 	}
 
-      if (*p && *p == '.')
-	{
-	  p++;
-	  fmt++;
-	  while (*p && grub_isdigit (*p))
-	    p++;
+      if (*fmt == '.')
+	fmt++;
+
+      if (grub_isdigit (*fmt))
+	format2 = grub_strtoul (fmt, (char **) &fmt, 10);
 
-	  if (p > fmt)
-	    {
-	      char fstr[p - fmt + 1];
-	      grub_strncpy (fstr, fmt, p - fmt);
-	      fstr[p - fmt] = 0;
-	      format2 = grub_strtoul (fstr, 0, 10);
-	      fmt = p;
-	    }
-	}
       if (*fmt == '$')
 	{
 	  curn = format1 - 1;
@@ -900,55 +876,58 @@ grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0, va_list a
 
       c = *fmt++;
       if (c == 'l')
+	c = *fmt++;
+      if (c == 'l')
+	c = *fmt++;
+
+      if (c == '%')
 	{
-	  longfmt = 1;
-	  c = *fmt++;
-	  if (c == 'l')
-	    {
-	      longlongfmt = 1;
-	      c = *fmt++;
-	    }
+	  write_char (str, &count, max_len,c);
+	  n--;
+	  continue;
 	}
 
-      if (curn >= count_args)
+      if (curn >= args->count)
 	continue;
 
+      long long curarg = args->ptr[curn].ll;
+
       switch (c)
 	{
 	case 'p':
-	  write_str ("0x");
+	  write_char (str, &count, max_len, '0');
+	  write_char (str, &count, max_len, 'x');
 	  c = 'x';
-	  longlongfmt |= (sizeof (void *) == sizeof (long long));
 	  /* Fall through. */
 	case 'x':
 	case 'u':
-	  unsig = 1;
-	  /* Fall through. */
 	case 'd':
-	  if (longlongfmt)
-	    grub_lltoa (tmp, c, args[curn].ll);
-	  else if (longfmt && unsig)
-	    grub_lltoa (tmp, c, (unsigned long) args[curn].l);
-	  else if (longfmt)
-	    grub_lltoa (tmp, c, args[curn].l);
-	  else if (unsig)
-	    grub_lltoa (tmp, c, (unsigned) args[curn].i);
-	  else
-	    grub_lltoa (tmp, c, args[curn].i);
-	  if (! rightfill && grub_strlen (tmp) < format1)
-	    write_fill (zerofill, format1 - grub_strlen (tmp));
-	  write_str (tmp);
-	  if (rightfill && grub_strlen (tmp) < format1)
-	    write_fill (zerofill, format1 - grub_strlen (tmp));
+	  {
+	    char tmp[32];
+	    const char *p = tmp;
+	    grub_size_t len;
+	    grub_size_t fill;
+
+	    len = grub_lltoa (tmp, c, curarg) - tmp;
+	    fill = len < format1 ? format1 - len : 0;
+	    if (! rightfill)
+	      while (fill--)
+		write_char (str, &count, max_len, zerofill);
+	    while (*p)
+	      write_char (str, &count, max_len, *p++);
+	    if (rightfill)
+	      while (fill--)
+		write_char (str, &count, max_len, zerofill);
+	  }
 	  break;
 
 	case 'c':
-	  write_char (args[curn].i & 0xff);
+	  write_char (str, &count, max_len,curarg & 0xff);
 	  break;
 
 	case 'C':
 	  {
-	    grub_uint32_t code = args[curn].w;
+	    grub_uint32_t code = curarg;
 	    int shift;
 	    unsigned mask;
 
@@ -967,21 +946,11 @@ grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0, va_list a
 		shift = 12;
 		mask = 0xe0;
 	      }
-	    else if (code <= 0x1fffff)
+	    else if (code <= 0x10ffff)
 	      {
 		shift = 18;
 		mask = 0xf0;
 	      }
-	    else if (code <= 0x3ffffff)
-	      {
-		shift = 24;
-		mask = 0xf8;
-	      }
-	    else if (code <= 0x7fffffff)
-	      {
-		shift = 30;
-		mask = 0xfc;
-	      }
 	    else
 	      {
 		code = '?';
@@ -989,44 +958,49 @@ grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt0, va_list a
 		mask = 0;
 	      }
 
-	    write_char (mask | (code >> shift));
+	    write_char (str, &count, max_len,mask | (code >> shift));
 
 	    for (shift -= 6; shift >= 0; shift -= 6)
-	      write_char (0x80 | (0x3f & (code >> shift)));
+	      write_char (str, &count, max_len,0x80 | (0x3f & (code >> shift)));
 	  }
 	  break;
 
 	case 's':
-	  p = args[curn].p;
-	  if (p)
-	    {
-	      grub_size_t len = 0;
-	      while (len < format2 && p[len])
-		len++;
+	  {
+	    grub_size_t len = 0;
+	    grub_size_t fill;
+	    const char *p = ((char *) (grub_addr_t) curarg) ? : "(null)";
+	    grub_size_t i;
 
-	      if (!rightfill && len < format1)
-		write_fill (zerofill, format1 - len);
+	    while (len < format2 && p[len])
+	      len++;
 
-	      grub_size_t i;
-	      for (i = 0; i < len; i++)
-		write_char (*p++);
+	    fill = len < format1 ? format1 - len : 0;
 
-	      if (rightfill && len < format1)
-		write_fill (zerofill, format1 - len);
-	    }
-	  else
-	    write_str ("(null)");
+	    if (!rightfill)
+	      while (fill--)
+		write_char (str, &count, max_len, zerofill);
+
+	    for (i = 0; i < len; i++)
+	      write_char (str, &count, max_len,*p++);
+
+	    if (rightfill)
+	      while (fill--)
+		write_char (str, &count, max_len, zerofill);
+	  }
 
 	  break;
 
 	default:
-	  write_char (c);
+	  write_char (str, &count, max_len,c);
 	  break;
 	}
     }
 
-  *str = '\0';
-
+  if (count < max_len)
+    str[count] = '\0';
+  else
+    str[max_len] = '\0';
   return count;
 }
 
@@ -1034,13 +1008,18 @@ int
 grub_vsnprintf (char *str, grub_size_t n, const char *fmt, va_list ap)
 {
   grub_size_t ret;
+  struct printf_args args;
 
   if (!n)
     return 0;
 
   n--;
 
-  ret = grub_vsnprintf_real (str, n, fmt, ap);
+  parse_printf_args (fmt, &args, ap);
+
+  ret = grub_vsnprintf_real (str, n, fmt, &args);
+
+  free_printf_args (&args);
 
   return ret < n ? ret : n;
 }
@@ -1063,22 +1042,26 @@ grub_xvasprintf (const char *fmt, va_list ap)
 {
   grub_size_t s, as = PREALLOC_SIZE;
   char *ret;
+  struct printf_args args;
+
+  parse_printf_args (fmt, &args, ap);
 
   while (1)
     {
-      va_list ap2;
       ret = grub_malloc (as + 1);
       if (!ret)
-	return NULL;
+	{
+	  free_printf_args (&args);
+	  return NULL;
+	}
 
-      va_copy (ap2, ap);
-
-      s = grub_vsnprintf_real (ret, as, fmt, ap2);
-
-      va_end (ap2);
+      s = grub_vsnprintf_real (ret, as, fmt, &args);
 
       if (s <= as)
-	return ret;
+	{
+	  free_printf_args (&args);
+	  return ret;
+	}
 
       grub_free (ret);
       as = s;
@@ -1099,7 +1082,7 @@ grub_xasprintf (const char *fmt, ...)
 }
 
 /* Abort GRUB. This function does not return.  */
-void
+static void __attribute__ ((noreturn))
 grub_abort (void)
 {
   grub_printf ("\nAborted.");
@@ -1115,20 +1098,19 @@ grub_abort (void)
   grub_exit ();
 }
 
-#if ! defined (__APPLE__) && !defined (GRUB_UTIL)
-/* GCC emits references to abort().  */
-void abort (void) __attribute__ ((alias ("grub_abort")));
-#endif
-
-#if NEED_REGISTER_FRAME_INFO && !defined(GRUB_UTIL)
-void __register_frame_info (void)
+void
+grub_fatal (const char *fmt, ...)
 {
-}
+  va_list ap;
 
-void __deregister_frame_info (void)
-{
+  va_start (ap, fmt);
+  grub_vprintf (_(fmt), ap);
+  va_end (ap);
+
+  grub_refresh ();
+
+  grub_abort ();
 }
-#endif
 
 #if BOOT_TIME_STATS
 
