@@ -24,6 +24,8 @@
 #include <grub/i386/pc/vesa_modes_table.h>
 #include <grub/i18n.h>
 
+#pragma GCC diagnostic ignored "-Wformat-nonliteral"
+
 struct legacy_command
 {
   const char *name;
@@ -70,13 +72,14 @@ static struct legacy_command legacy_commands[] =
      "Print the blocklist notation of the file FILE."},
     {"boot", "boot\n", NULL, 0, 0, {}, 0, 0,
      "Boot the OS/chain-loader which has been loaded."},
-    {"bootp", "net_bootp; net_ls_addr; if [ x%s = x--with-configfile ]; then "
+    {"bootp", "net_bootp; net_ls_addr; echo $\"" N_("Default server is ${net_default_server}") "\"; if [ x%s = x--with-configfile ]; then "
      "if net_get_dhcp_option configfile_name pxe 150 string; then "
      "configfile $configfile_name; fi; fi\n", NULL, 0, 1,
      {TYPE_WITH_CONFIGFILE_OPTION}, FLAG_IGNORE_REST, "[--with-configfile]",
      "Initialize a network device via BOOTP. If the option `--with-configfile'"
      " is given, try to load a configuration file specified by the 150 vendor"
      " tag."},
+    /* FIXME: border unsupported.  */
     {"cat", "cat '%s'\n", NULL, 0, 1, {TYPE_FILE}, 0, "FILE",
      "Print the contents of the file FILE."},
     {"chainloader", "chainloader %s '%s'\n", NULL, 0,
@@ -249,6 +252,9 @@ static struct legacy_command legacy_commands[] =
     {"pause", "echo %s; if ! sleep -i 60; then return; fi\n", NULL, 0, 1,
      {TYPE_REST_VERBATIM}, 0,
      "[MESSAGE ...]", "Print MESSAGE, then wait until a key is pressed."},
+    {"print", "echo %s\n", NULL, 0, 1,
+     {TYPE_REST_VERBATIM}, 0,
+     "[MESSAGE ...]", "Print MESSAGE."},
     /* FIXME: quit unsupported.  */
     /* FIXME: rarp unsupported.  */
     {"read", "read_dword %s\n", NULL, 0, 1, {TYPE_INT}, 0, "ADDR",
@@ -295,6 +301,7 @@ static struct legacy_command legacy_commands[] =
      " STOP is the length of stop bit(s). The option --device can be used only"
      " in the grub shell, which specifies the file name of a tty device. The"
      " default values are COM1, 9600, 8N1."},
+    /* FIXME: shade unsupported.  */
     /* FIXME: silent unsupported.  */
     /* FIXME: splashimage unsupported.  */
     /* FIXME: setkey unsupported.  */    /* NUL_TERMINATE */
@@ -305,7 +312,7 @@ static struct legacy_command legacy_commands[] =
     /* FIXME: graphics unsupported.  */
     {"terminal", NULL, NULL, 0, 0, {}, FLAG_TERMINAL | FLAG_IGNORE_REST,
      "[--dumb] [--no-echo] [--no-edit] [--timeout=SECS] [--lines=LINES] "
-     "[--silent] [console] [serial] [hercules]",
+     "[--silent] [console] [serial] [hercules] [graphics]",
      "Select a terminal. When multiple terminals are specified, wait until"
      " you push any key to continue. If both console and serial are specified,"
      " the terminal to which you input a key first will be selected. If no"
@@ -346,6 +353,7 @@ static struct legacy_command legacy_commands[] =
      FLAG_FALLBACK, NULL, NULL}
     /* FIXME: verbose unsupported.  */
     /* FIXME: version unsupported.  */
+    /* FIXME: viewport unsupported.  */
   };
 
 char *
@@ -385,6 +393,29 @@ adjust_file (const char *in, grub_size_t len)
     if (*ptr == '\'' || *ptr == '\\')
       overhead++;
   comma = ptr;
+  if (*comma == ')' && comma - in == 3
+      && in[1] == 'n' && in[2] == 'd')
+    {
+      rest = comma + 1;
+      for (ptr = rest; ptr < in + len && *ptr; ptr++)
+	if (*ptr == '\'' || *ptr == '\\')
+	  overhead++;
+
+      ret = grub_malloc (ptr - in + overhead + 15);
+      if (!ret)
+	return NULL;
+
+      outptr = grub_stpcpy (ret, "(tftp)");;
+      for (ptr = rest; ptr < in + len; ptr++)
+	{
+	  if (*ptr == '\'' || *ptr == '\\')
+	    *outptr++ = '\\';
+
+	  *outptr++ = *ptr;
+	}
+      *outptr = 0;
+      return ret;
+    }
   if (*comma != ',')
     return grub_legacy_escape (in, len);
   part = grub_strtoull (comma + 1, (char **) &rest, 0);
@@ -398,7 +429,7 @@ adjust_file (const char *in, grub_size_t len)
       overhead++;
 
   /* 35 is enough for any 2 numbers.  */
-  ret = grub_malloc (ptr - in + overhead + 35);
+  ret = grub_malloc (ptr - in + overhead + 35 + 5);
   if (!ret)
     return NULL;
 
@@ -521,9 +552,9 @@ grub_legacy_parse (const char *buf, char **entryname, char **suffix)
       int no_echo = 0, no_edit = 0;
 #endif
       int hercules = 0;
-      int console = 0, serial = 0;
+      int console = 0, serial = 0, graphics = 0;
       /* Big enough for any possible resulting command. */
-      char outbuf[256] = "";
+      char outbuf[512] = "";
       char *outptr;
       while (*ptr)
 	{
@@ -555,46 +586,38 @@ grub_legacy_parse (const char *buf, char **entryname, char **suffix)
 	    serial = 1;
 	  if (grub_memcmp (ptr, "hercules", sizeof ("hercules") - 1) == 0)
 	    hercules = 1;
+	  if (grub_memcmp (ptr, "graphics", sizeof ("graphics") - 1) == 0)
+	    graphics = 1;
 	  while (*ptr && !grub_isspace (*ptr))
 	    ptr++;
 	  while (*ptr && grub_isspace (*ptr))
 	    ptr++;
 	}
 
-      if (!console && !serial)
+      if (!console && !serial && !hercules && !graphics)
 	return grub_strdup ("terminal_input; terminal_output; terminfo\n");
 
-      grub_strcpy (outbuf, "terminal_input ");
-      outptr = outbuf + grub_strlen (outbuf);
+      outptr = outbuf;
+
+      if (graphics)
+	outptr = grub_stpcpy (outptr, "insmod all_video; ");
+
+      outptr = grub_stpcpy (outptr, "terminal_input ");
       if (serial)
-	{
-	  grub_strcpy (outptr, "serial ");
-	  outptr += grub_strlen (outptr);
-	}
-      if (console || hercules)
-	{
-	  grub_strcpy (outptr, "console ");
-	  outptr += grub_strlen (outptr);
-	}
-      grub_strcpy (outptr, "; terminal_output ");
-      outptr += grub_strlen (outptr);
+	outptr = grub_stpcpy (outptr, "serial ");
+      if (console || hercules || graphics)
+	outptr = grub_stpcpy (outptr, "console ");
+      outptr = grub_stpcpy (outptr, "; terminal_output ");
       if (serial)
-	{
-	  grub_strcpy (outptr, "serial ");
-	  outptr += grub_strlen (outptr);
-	}
+	outptr = grub_stpcpy (outptr, "serial ");
       if (console)
-	{
-	  grub_strcpy (outptr, "console ");
-	  outptr += grub_strlen (outptr);
-	}
+	outptr = grub_stpcpy (outptr, "console ");
       if (hercules)
-	{
-	  grub_strcpy (outptr, "mda_text ");
-	  outptr += grub_strlen (outptr);
-	}
-      grub_strcpy (outptr, "; ");
-      outptr += grub_strlen (outptr);
+	outptr = grub_stpcpy (outptr, "mda_text ");
+      if (graphics)
+	outptr = grub_stpcpy (outptr, "gfxterm ");
+      outptr = grub_stpcpy (outptr, "; ");
+      *outptr = '\0';
       if (serial)
 	{
 	  grub_snprintf (outptr, outbuf + sizeof (outbuf) - outptr,
@@ -641,6 +664,7 @@ grub_legacy_parse (const char *buf, char **entryname, char **suffix)
 	  {
 	  case TYPE_FILE_NO_CONSUME:
 	    hold_arg = 1;
+	    /* Fallthrough.  */
 	  case TYPE_PARTITION:
 	  case TYPE_FILE:
 	    args[i] = adjust_file (curarg, curarglen);
