@@ -34,6 +34,12 @@ struct dns_cache_element
 #define DNS_CACHE_SIZE 1021
 #define DNS_HASH_BASE 423
 
+typedef enum grub_dns_qtype_id
+  {
+    GRUB_DNS_QTYPE_A = 1,
+    GRUB_DNS_QTYPE_AAAA = 28
+  } grub_dns_qtype_id_t;
+
 static struct dns_cache_element dns_cache[DNS_CACHE_SIZE];
 static struct grub_net_network_level_address *dns_servers;
 static grub_size_t dns_nservers, dns_servers_alloc;
@@ -80,7 +86,7 @@ struct dns_header
   grub_uint16_t ancount;
   grub_uint16_t nscount;
   grub_uint16_t arcount;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 enum
   {
@@ -114,7 +120,7 @@ struct recv_data
 static inline int
 hash (const char *str)
 {
-  int v = 0, xn = 1;
+  unsigned v = 0, xn = 1;
   const char *ptr;
   for (ptr = str; *ptr; )
     {
@@ -232,6 +238,15 @@ recv_hook (grub_net_udp_socket_t sock __attribute__ ((unused)),
   char *redirect_save = NULL;
   grub_uint32_t ttl_all = ~0U;
 
+  /* Code apparently assumed that only one packet is received as response.
+     We may get multiple responses due to network condition, so check here
+     and quit early. */
+  if (*data->addresses)
+    {
+      grub_netbuff_free (nb);
+      return GRUB_ERR_NONE;
+    }
+
   head = (struct dns_header *) nb->data;
   ptr = (grub_uint8_t *) (head + 1);
   if (ptr >= nb->tail)
@@ -256,7 +271,7 @@ recv_hook (grub_net_udp_socket_t sock __attribute__ ((unused)),
       grub_netbuff_free (nb);
       return GRUB_ERR_NONE;
     }
-  for (i = 0; i < grub_cpu_to_be16 (head->qdcount); i++)
+  for (i = 0; i < grub_be_to_cpu16 (head->qdcount); i++)
     {
       if (ptr >= nb->tail)
 	{
@@ -271,7 +286,7 @@ recv_hook (grub_net_udp_socket_t sock __attribute__ ((unused)),
       ptr += 4;
     }
   *data->addresses = grub_malloc (sizeof ((*data->addresses)[0])
-				 * grub_cpu_to_be16 (head->ancount));
+				 * grub_be_to_cpu16 (head->ancount));
   if (!*data->addresses)
     {
       grub_errno = GRUB_ERR_NONE;
@@ -280,7 +295,7 @@ recv_hook (grub_net_udp_socket_t sock __attribute__ ((unused)),
     }
   reparse_ptr = ptr;
  reparse:
-  for (i = 0, ptr = reparse_ptr; i < grub_cpu_to_be16 (head->ancount); i++)
+  for (i = 0, ptr = reparse_ptr; i < grub_be_to_cpu16 (head->ancount); i++)
     {
       int ignored = 0;
       grub_uint8_t class;
@@ -421,16 +436,17 @@ grub_net_dns_lookup (const char *name,
   grub_size_t send_servers = 0;
   grub_size_t i, j;
   struct grub_net_buff *nb;
-  grub_net_udp_socket_t sockets[n_servers];
+  grub_net_udp_socket_t *sockets;
   grub_uint8_t *optr;
   const char *iptr;
   struct dns_header *head;
   static grub_uint16_t id = 1;
+  grub_uint8_t *qtypeptr;
   grub_err_t err = GRUB_ERR_NONE;
   struct recv_data data = {naddresses, addresses, cache,
 			   grub_cpu_to_be16 (id++), 0, 0, name, 0};
   grub_uint8_t *nbd;
-  int have_server = 0;
+  grub_size_t try_server = 0;
 
   if (!servers)
     {
@@ -463,18 +479,25 @@ grub_net_dns_lookup (const char *name,
 	}
     }
 
+  sockets = grub_malloc (sizeof (sockets[0]) * n_servers);
+  if (!sockets)
+    return grub_errno;
+
   data.name = grub_strdup (name);
   if (!data.name)
-    return grub_errno;
+    {
+      grub_free (sockets);
+      return grub_errno;
+    }
 
   nb = grub_netbuff_alloc (GRUB_NET_OUR_MAX_IP_HEADER_SIZE
 			   + GRUB_NET_MAX_LINK_HEADER_SIZE
 			   + GRUB_NET_UDP_HEADER_SIZE
 			   + sizeof (struct dns_header)
-			   + grub_strlen (name) + 2 + 4
-			   + 2 + 4);
+			   + grub_strlen (name) + 2 + 4);
   if (!nb)
     {
+      grub_free (sockets);
       grub_free (data.name);
       return grub_errno;
     }
@@ -482,7 +505,7 @@ grub_net_dns_lookup (const char *name,
 			+ GRUB_NET_MAX_LINK_HEADER_SIZE
 			+ GRUB_NET_UDP_HEADER_SIZE);
   grub_netbuff_put (nb, sizeof (struct dns_header)
-		    + grub_strlen (name) + 2 + 4 + 2 + 4);
+		    + grub_strlen (name) + 2 + 4);
   head = (struct dns_header *) nb->data;
   optr = (grub_uint8_t *) (head + 1);
   for (iptr = name; *iptr; )
@@ -493,6 +516,7 @@ grub_net_dns_lookup (const char *name,
 	dot = iptr + grub_strlen (iptr);
       if ((dot - iptr) >= 64)
 	{
+	  grub_free (sockets);
 	  grub_free (data.name);
 	  return grub_error (GRUB_ERR_BAD_ARGUMENT,
 			     N_("domain name component is too long"));
@@ -507,20 +531,9 @@ grub_net_dns_lookup (const char *name,
     }
   *optr++ = 0;
 
-  /* Type: A.  */
+  /* Type.  */
   *optr++ = 0;
-  *optr++ = 1;
-
-  /* Class.  */
-  *optr++ = 0;
-  *optr++ = 1;
-
-  /* Compressed name.  */
-  *optr++ = 0xc0;
-  *optr++ = 0x0c;
-  /* Type: AAAA.  */
-  *optr++ = 0;
-  *optr++ = 28;
+  qtypeptr = optr++;
 
   /* Class.  */
   *optr++ = 0;
@@ -529,7 +542,7 @@ grub_net_dns_lookup (const char *name,
   head->id = data.id;
   head->flags = FLAGS_RD;
   head->ra_z_r_code = 0;
-  head->qdcount = grub_cpu_to_be16_compile_time (2);
+  head->qdcount = grub_cpu_to_be16_compile_time (1);
   head->ancount = grub_cpu_to_be16_compile_time (0);
   head->nscount = grub_cpu_to_be16_compile_time (0);
   head->arcount = grub_cpu_to_be16_compile_time (0);
@@ -539,42 +552,54 @@ grub_net_dns_lookup (const char *name,
   for (i = 0; i < n_servers * 4; i++)
     {
       /* Connect to a next server.  */
-      while (!(i & 1) && send_servers < n_servers)
+      while (!(i & 1) && try_server < n_servers)
 	{
-	  sockets[send_servers] = grub_net_udp_open (servers[send_servers],
+	  sockets[send_servers] = grub_net_udp_open (servers[try_server++],
 						     DNS_PORT,
 						     recv_hook,
 						     &data);
-	  send_servers++;
-	  if (!sockets[send_servers - 1])
+	  if (!sockets[send_servers])
 	    {
 	      err = grub_errno;
 	      grub_errno = GRUB_ERR_NONE;
 	    }
 	  else
 	    {
-	      have_server = 1;
+	      send_servers++;
 	      break;
 	    }
 	}
-      if (!have_server)
+      if (!send_servers)
 	goto out;
       if (*data.naddresses)
 	goto out;
       for (j = 0; j < send_servers; j++)
 	{
-	  grub_err_t err2;
-	  if (!sockets[j])
-	    continue;
-	  nb->data = nbd;
-	  err2 = grub_net_send_udp_packet (sockets[j], nb);
-	  if (err2)
-	    {
-	      grub_errno = GRUB_ERR_NONE;
-	      err = err2;
-	    }
-	  if (*data.naddresses)
-	    goto out;
+          grub_err_t err2;
+
+          grub_size_t t = 0;
+          do
+            {
+              nb->data = nbd;
+              if (servers[j].option == DNS_OPTION_IPV4 ||
+                 ((servers[j].option == DNS_OPTION_PREFER_IPV4) && (t++ == 0)) ||
+                 ((servers[j].option == DNS_OPTION_PREFER_IPV6) && (t++ == 1)))
+                *qtypeptr = GRUB_DNS_QTYPE_A;
+              else
+                *qtypeptr = GRUB_DNS_QTYPE_AAAA;
+
+              grub_dprintf ("dns", "QTYPE: %u QNAME: %s\n", *qtypeptr, name);
+
+              err2 = grub_net_send_udp_packet (sockets[j], nb);
+              if (err2)
+                {
+                  grub_errno = GRUB_ERR_NONE;
+                  err = err2;
+                }
+              if (*data.naddresses)
+                goto out;
+            }
+          while (t == 1);
 	}
       grub_net_poll_cards (200, &data.stop);
     }
@@ -584,6 +609,8 @@ grub_net_dns_lookup (const char *name,
   for (j = 0; j < send_servers; j++)
     grub_net_udp_close (sockets[j]);
   
+  grub_free (sockets);
+
   if (*data.naddresses)
     return GRUB_ERR_NONE;
   if (data.dns_err)
@@ -604,22 +631,28 @@ grub_cmd_nslookup (struct grub_command *cmd __attribute__ ((unused)),
 		   int argc, char **args)
 {
   grub_err_t err;
-  grub_size_t naddresses, i;
+  struct grub_net_network_level_address cmd_server;
+  struct grub_net_network_level_address *servers;
+  grub_size_t nservers, i, naddresses = 0;
   struct grub_net_network_level_address *addresses = 0;
   if (argc != 2 && argc != 1)
     return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("two arguments expected"));
   if (argc == 2)
     {
-      struct grub_net_network_level_address server;
-      err = grub_net_resolve_address (args[1], &server);
+      err = grub_net_resolve_address (args[1], &cmd_server);
       if (err)
 	return err;
-      err = grub_net_dns_lookup (args[0], &server, 1, &naddresses,
-				 &addresses, 0);
+      servers = &cmd_server;
+      nservers = 1;
     }
   else
-    err = grub_net_dns_lookup (args[0], dns_servers, dns_nservers, &naddresses,
-			       &addresses, 0);
+    {
+      servers = dns_servers;
+      nservers = dns_nservers;
+    }
+
+  grub_net_dns_lookup (args[0], servers, nservers, &naddresses,
+                       &addresses, 0);
 
   for (i = 0; i < naddresses; i++)
     {
@@ -628,7 +661,9 @@ grub_cmd_nslookup (struct grub_command *cmd __attribute__ ((unused)),
       grub_printf ("%s\n", buf);
     }
   grub_free (addresses);
-  return GRUB_ERR_NONE;
+  if (naddresses)
+    return GRUB_ERR_NONE;
+  return grub_error (GRUB_ERR_NET_NO_DOMAIN, N_("no DNS record found"));
 }
 
 static grub_err_t
@@ -637,11 +672,32 @@ grub_cmd_list_dns (struct grub_command *cmd __attribute__ ((unused)),
 		   char **args __attribute__ ((unused)))
 {
   grub_size_t i;
+  const char *strtype = "";
+
   for (i = 0; i < dns_nservers; i++)
     {
+      switch (dns_servers[i].option)
+        {
+        case DNS_OPTION_IPV4:
+          strtype = _("only ipv4");
+          break;
+
+        case DNS_OPTION_IPV6:
+          strtype = _("only ipv6");
+          break;
+
+        case DNS_OPTION_PREFER_IPV4:
+          strtype = _("prefer ipv4");
+          break;
+
+        case DNS_OPTION_PREFER_IPV6:
+          strtype = _("prefer ipv6");
+          break;
+        }
+
       char buf[GRUB_NET_MAX_STR_ADDR_LEN];
       grub_net_addr_to_str (&dns_servers[i], buf);
-      grub_printf ("%s\n", buf);
+      grub_printf ("%s (%s)\n", buf, strtype);
     }
   return GRUB_ERR_NONE;
 }
@@ -653,8 +709,24 @@ grub_cmd_add_dns (struct grub_command *cmd __attribute__ ((unused)),
   grub_err_t err;
   struct grub_net_network_level_address server;
 
-  if (argc != 1)
+  if ((argc < 1) || (argc > 2))
     return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("one argument expected"));
+  else if (argc == 1)
+    server.option = DNS_OPTION_PREFER_IPV4;
+  else
+    {
+      if (grub_strcmp (args[1], "--only-ipv4") == 0)
+          server.option = DNS_OPTION_IPV4;
+      else if (grub_strcmp (args[1], "--only-ipv6") == 0)
+          server.option = DNS_OPTION_IPV6;
+      else if (grub_strcmp (args[1], "--prefer-ipv4") == 0)
+          server.option = DNS_OPTION_PREFER_IPV4;
+      else if (grub_strcmp (args[1], "--prefer-ipv6") == 0)
+          server.option = DNS_OPTION_PREFER_IPV6;
+      else
+        return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("invalid argument"));
+    }
+
   err = grub_net_resolve_address (args[0], &server);
   if (err)
     return err;
