@@ -102,10 +102,10 @@ cryptolist.write ("CRC64: crc64\n");
 for cipher_file in cipher_files:
     infile = os.path.join (cipher_dir_in, cipher_file)
     outfile = os.path.join (cipher_dir_out, cipher_file)
-    if cipher_file == "ChangeLog":
+    if cipher_file == "ChangeLog" or cipher_file == "ChangeLog-2011":
         continue
     chlognew = "	* %s" % cipher_file
-    if re.match ("(Manifest|Makefile\.am|ac\.c|cipher\.c|hash-common\.c|hmac-tests\.c|md\.c|pubkey\.c)$", cipher_file) or cipher_file == "elgamal.c" or cipher_file == "primegen.c" or cipher_file == "test-getrusage.c":
+    if re.match ("(Manifest|Makefile\.am|ac\.c|cipher\.c|hash-common\.c|hmac-tests\.c|md\.c|pubkey\.c)$", cipher_file) or cipher_file == "kdf.c" or cipher_file == "elgamal.c" or cipher_file == "primegen.c" or cipher_file == "ecc.c" or cipher_file == "test-getrusage.c":
         chlog = "%s%s: Removed\n" % (chlog, chlognew)
         continue
     # Autogenerated files. Not even worth mentionning in ChangeLog
@@ -137,11 +137,13 @@ for cipher_file in cipher_files:
 
         ciphernames = []
         mdnames = []
+        mdctxsizes = []
         pknames = []
         hold = False
-        skip = False
+        skip = 0
         skip2 = False
         ismd = False
+        mdarg = 0
         ispk = False
         iscipher = False
         iscryptostart = False
@@ -160,9 +162,9 @@ for cipher_file in cipher_files:
                 if not re.search (";", line) is None:
                     skip_statement = False
                 continue
-            if skip:
+            if skip > 0:
                 if line[0] == "}":
-                    skip = False
+                    skip = skip - 1
                 continue
             if skip2:
                 if not re.search (" *};", line) is None:
@@ -174,6 +176,11 @@ for cipher_file in cipher_files:
                     sg = s.groups()[0]
                     cryptolist.write (("%s: %s\n") % (sg, modname))
                     iscryptostart = False
+            if ismd:
+                spl = line.split (",")
+                if mdarg + len (spl) > 9 and mdarg <= 9 and ("sizeof" in spl[9-mdarg]):
+                    mdctxsizes.append (spl[9-mdarg].lstrip ().rstrip())
+                mdarg = mdarg + len (spl) - 1
             if ismd or iscipher or ispk:
                 if not re.search (" *};", line) is None:
                     if not iscomma:
@@ -189,13 +196,14 @@ for cipher_file in cipher_files:
                         fw.write ("    .blocksize = %s\n"
                                   % mdblocksizes [mdname])
                     ismd = False
+                    mdarg = 0
                     iscipher = False
                     ispk = False
                 iscomma = not re.search (",$", line) is None
             # Used only for selftests.
             m = re.match ("(static byte|static unsigned char) (weak_keys_chksum)\[[0-9]*\] =", line)
             if not m is None:
-                skip = True
+                skip = 1
                 fname = m.groups ()[1]
                 chmsg = "(%s): Removed." % fname
                 if nch:
@@ -209,7 +217,11 @@ for cipher_file in cipher_files:
                 # We're optimising for size and exclude anything needing good
                 # randomness.
                 if not re.match ("(run_selftests|selftest|_gcry_aes_c.._..c|_gcry_[a-z0-9]*_hash_buffer|tripledes_set2keys|do_tripledes_set_extra_info|_gcry_rmd160_mixblock|serpent_test|dsa_generate_ext|test_keys|gen_k|sign|gen_x931_parm_xp|generate_x931|generate_key|dsa_generate|dsa_sign|ecc_sign|generate|generate_fips186|_gcry_register_pk_dsa_progress|_gcry_register_pk_ecc_progress|progress|scanval|ec2os|ecc_generate_ext|ecc_generate|compute_keygrip|ecc_get_param|_gcry_register_pk_dsa_progress|gen_x931_parm_xp|gen_x931_parm_xi|rsa_decrypt|rsa_sign|rsa_generate_ext|rsa_generate|secret|check_exponent|rsa_blind|rsa_unblind|extract_a_from_sexp|curve_free|curve_copy|point_set)", line) is None:
-                    skip = True
+
+                    skip = 1
+                    if not re.match ("selftest", line) is None and cipher_file == "idea.c":
+                        skip = 3
+
                     if not re.match ("serpent_test", line) is None:
                         fw.write ("static const char *serpent_test (void) { return 0; }\n");
                     if not re.match ("dsa_generate", line) is None:
@@ -279,6 +291,7 @@ for cipher_file in cipher_files:
                 mdname = re.match("[a-zA-Z0-9_]*",mdname).group ()
                 mdnames.append (mdname)
                 ismd = True
+                mdarg = 0
                 iscryptostart = True
             m = re.match ("static const char \*selftest.*;$", line)
             if not m is None:
@@ -343,9 +356,24 @@ for cipher_file in cipher_files:
                 holdline = line
                 continue
             m = re.match ("static int tripledes_set2keys \(.*\);", line)
+            if not m is None:
+                continue
+            m = re.match ("static int tripledes_set3keys \(.*\);", line)
             if not m is None:
                 continue
             m = re.match ("static int tripledes_set2keys \(", line)
+            if not m is None:
+                skip_statement = True
+                continue
+            m = re.match ("static int tripledes_set3keys \(", line)
+            if not m is None:
+                skip_statement = True
+                continue
+            m = re.match ("static const char sample_secret_key", line)
+            if not m is None:
+                skip_statement = True
+                continue
+            m = re.match ("static const char sample_public_key", line)
             if not m is None:
                 skip_statement = True
                 continue
@@ -412,6 +440,8 @@ for cipher_file in cipher_files:
                 chmsg = "Register cipher %s" % ciphername
                 chlognew = "%s\n	%s" % (chlognew, chmsg)
                 fw.write ("  grub_cipher_register (&%s);\n" % ciphername)
+            for ctxsize in mdctxsizes:
+                fw.write ("  COMPILE_TIME_ASSERT(%s <= GRUB_CRYPTO_MAX_MD_CONTEXT_SIZE);\n" % ctxsize)
             for mdname in mdnames:
                 chmsg = "Register digest %s" % mdname
                 chlognew = "%s\n	%s" % (chlognew, chmsg)
@@ -451,7 +481,7 @@ for cipher_file in cipher_files:
                 conf.write ("  cflags = '$(CFLAGS_GCRY) -Wno-redundant-decls -Wno-sign-compare';\n")
             elif modname == "gcry_rijndael" or modname == "gcry_md4" or modname == "gcry_md5" or modname == "gcry_rmd160" or modname == "gcry_sha1" or modname == "gcry_sha256" or modname == "gcry_sha512" or modname == "gcry_tiger":
                 # Alignment checked by hand
-                conf.write ("  cflags = '$(CFLAGS_GCRY) -Wno-cast-align -Wno-strict-aliasing';\n");
+                conf.write ("  cflags = '$(CFLAGS_GCRY) -Wno-cast-align';\n");
             else:
                 conf.write ("  cflags = '$(CFLAGS_GCRY)';\n");
             conf.write ("  cppflags = '$(CPPFLAGS_GCRY)';\n");
@@ -494,7 +524,7 @@ for src in sorted (os.listdir (os.path.join (indir, "src"))):
         fw.close ()
         continue
     if src == "visibility.h":
-        fw.write ("# include <grub/gcrypt/gcrypt.h>")
+        fw.write ("# include <grub/gcrypt/gcrypt.h>\n")
         fw.close ()
         continue
     f = codecs.open (infile, "r", "utf-8")
@@ -504,6 +534,12 @@ for src in sorted (os.listdir (os.path.join (indir, "src"))):
         fw.close ()
         continue
 
+    if src == "g10lib.h":
+        fw.write (f.read ().replace ("(printf,f,a)", "(__printf__,f,a)"))
+        f.close ()
+        fw.close ()
+        continue
+
     fw.write (f.read ())
     f.close ()
     fw.close ()
@@ -522,18 +558,18 @@ for src in sorted (os.listdir (os.path.join (indir, "mpi"))):
     fw.write ("/* This file was automatically imported with \n")
     fw.write ("   import_gcry.py. Please don't modify it */\n")
     hold = False
-    skip = False
+    skip = 0
     for line in f:
-        if skip:
+        if skip > 0:
             if line[0] == "}":
-                skip = False
+                skip = skip - 1
             continue
         if hold:
             hold = False
             # We're optimising for size and exclude anything needing good
             # randomness.
             if not re.match ("(_gcry_mpi_get_hw_config|gcry_mpi_randomize)", line) is None:
-                skip = True
+                skip = 1
                 continue
             else:
                 fw.write (holdline)
